Unit DirView;

{===============================================================
 Component TDirView / Version 2.4, July 1999
 ===============================================================


    Description:
    ============
    Displays files of a single directory as listview with shell
    icons. Complete drag&Drop support for files and directories.


    Author:
    =======
    (c) Ingo Eckel 1998, 1999
    Sodener Weg 38
    65812 Bad Soden
    Germany

    For detailed documentation and history see TDirView.htm.

    V2.4:
    - New function GetShellDisplayName
    - Using the displayname of deleted files in the recycle bin.
    - Display the contextmenu and property sheet even in the recycle bin.
    - Perform drag&drop with items from the recycle bin as source.
    - Flickering column header fixed when painting or scrolling.
    - Bug fixed when displaying the context menu: after this, filesystem changes
      where not detected any more.
    - Bug fixed when ViewStyle = vsList and property UseIconUpdateThread is
      set to True: the icons have been not updated.
    - New method FindFileItem (former: FindCaption)

{===============================================================}

{ Define DRAGDROP to enable drag&drop for files and directories:}
{$DEFINE DRAGDROP}

{ Define GRAYCOLICON to display the column icon without colors, => only gray:}
{.$DEFINE GRAYCOLICON}

{Required compiler options for TDirView:}
{$A+,B-,X+,H+,P+}

{===============================================================}
interface
{===============================================================}
uses
  Windows, Messages, SysUtils,
  Classes, Graphics, Controls,
  Forms, Dialogs, ComCtrls,
  ShellAPI, Math, ComObj, ShlObj,
  ActiveX, CommCtrl, Extctrls, ImgList,
  PIDL,
  ListExt,
  FileOperator,
  DiscMon,
  ShellDialogs,
  IEDriveInfo,
{$IFDEF DRAGDROP}
  DragDrop,
  DragDropFilesEx,
  FileChanges,
{$ENDIF}
  BaseUtils;


{$I ResStrings.pas }
{$R ColIcons.res}


Const msThreadChangeDelay = 10; {TDiscMonitor: change delay}
      MaxWaitTimeOut      = 10; {TFileDeleteThread: wait nn seconds for deleting files or directories}
      SpecialExtensions   = 'EXE,LNK,ICO,ANI,CUR,PIF,JOB,CPL';
      ExeExtension        = 'EXE';
      MinDate             = $21;    {01.01.1980}
      MaxDate             = $EF9F;  {31.12.2099}
      MinTime             = 0;      {00:00:00}
      MaxTime             = $C000;  {24:00:00}
      FileAttr            = faAnyFile and Not (faVolumeID{ Or faDirectory});
      ColCount            = 5;      {Five colums: 0..4}
      AnimateDelay        = 1000;    {Delay im ms when displaying the animation}


type
  {Exceptions:}
  EIUThread           = class(Exception);
  EDragDrop           = class(Exception);
  EExecFile           = class(Exception);
  EInvalidFileName    = class(Exception);
  ERenameFileFailed   = class(Exception);

  TSortColumn         = (sName, sSize, sType, sDate, sAttr);
  TSelAttr            = (SelDontCare, SelYes, SelNo);
  TDateTimeDisplay    = (dtdDateTimeSec, dtdDateTime, dtdDate);
  TClipBoardOperation = (cboNone, cboCut, cboCopy);
  {Drag&drop errors:}
  TDDError            = (DDCreateShortCutError, DDPathNotFoundError);

  ExtStr = String[4];

  TIntArray = array of Integer;
  PIntArray = ^TIntArray;
  TColumnOrder = Array[0..ColCount-1] of Integer;

  {Record for each file item:}
  PFileRec = ^TFileRec;
  TFileRec = record
    Empty       : Boolean;
    IconEmpty   : Boolean;
    isDirectory : Boolean;
    isRecycleBin: Boolean;
    FileName    : String;
    Displayname : String;
    FileExt     : ExtStr;
    TypeName    : String;
    ImageIndex  : Integer;
    Size        : Cardinal;
    Attr        : DWORD;
    FileTime    : TFILETIME;
    PIDL        : PItemIDList;
  end;

  {Record for fileinfo caching:}
  PInfoCache = ^TInfoCache;
  TInfoCache = record
    FileExt     : ExtStr;
    TypeName    : ShortString;
    ImageIndex  : Integer;
  end;

  {Record for icon-caching (Procedure Reload):}
  PIconRec = ^TIconRec;
  TIconRec = Record
    FileName   : String[125];
    ImageIndex : Integer;
  End;

  {Additional events:}
  TDirViewAddFileEvent   = Procedure(Sender: TObject; Var SearchRec: TSearchRec; Var AddFile : Boolean) of object;
  TDirViewExecFileEvent  = Procedure(Sender: TObject; Item : TListItem; Var AllowExec: Boolean) of object;

{$IFDEF DRAGDROP}
  {Drag&Drop events:}
  TDDOnDragEnter         = Procedure(Sender: TObject; DataObj: IDataObject; grfKeyState: Longint; pt: TPoint; var dwEffect: longint; var Accept:boolean) of object;
  TDDOnDragLeave         = Procedure(Sender: TObject) of object;
  TDDOnDragOver          = Procedure(Sender: TObject; grfKeyState: Longint; pt: TPoint; var dwEffect: longint) of object;
  TDDOnDrop              = Procedure(Sender: TObject; DataObj: IDataObject; grfKeyState: Longint;  pt: TPoint; var dwEffect: longint) of object;
  TDDOnQueryContinueDrag = Procedure(Sender: TObject; fEscapePressed: BOOL; grfKeyState: Longint; var Result: HResult) of object;
  TDDOnGiveFeedback      = Procedure(Sender: TObject; dwEffect: Longint; var Result: HResult) of object;
  TDDOnDragDetect        = Procedure(Sender: TObject; grfKeyState: Longint; DetectStart, Pt: TPoint; DragStatus:TDragDetectStatus) of object;

  TDDErrorEvent          = Procedure(Sender: TObject; ErrorNo : TDDError) of Object;
  TDDExecutedEvent       = Procedure(Sender: TObject; dwEffect: Longint) of Object;
  TDDFileOperationEvent  = Procedure(Sender: TObject; dwEffect: LongInt; SourcePath, TargetPath: String;
                                    Var DoOperation : Boolean) of Object;
  TDDFileOperationExecutedEvent = Procedure(Sender: TObject; dwEffect: LongInt; SourcePath, TargetPath: String) of Object;
{$ENDIF}

  {Interface record for procedure SelectFiles:}
  TFileFilter = Record
    Mask            : String;
    SelArchive      : TSelAttr;
    SelHidden       : TSelAttr;
    SelSysFile      : TSelAttr;
    SelReadOnly     : TSelAttr;
    SelFileSizeFrom : Integer;
    SelFileSizeTo   : Integer;
    SelFileDateFrom : Word;
    SelFileDateTo   : Word;
    SelFileTimeFrom : Word;
    SelFileTimeTo   : Word;
  End;

  TDirView = Class;

{===============================================================}
{  TIconUpdateThread:                                           }
{  Fetch shell icons via thread.                                }
{===============================================================}

{---------------------------------------------------------------}
  TIconUpdateThread = class(TThread)
{---------------------------------------------------------------}
  private
{---------------------------------------------------------------}
    { Private-Deklarationen}
    fOwner       : TDirView;
    fIndex       : Integer;
    fMaxIndex    : Integer;
    fNewIcons    : Boolean;
    fSyncIcon    : Integer;
    CurrentIndex : Integer;
    CurrentFilePath : String;
    CurrentItemData : TFileRec;
    InvalidItem  : Boolean;

    Procedure SetIndex(i : Integer);
    Procedure SetMaxIndex(i : Integer);

{---------------------------------------------------------------}
  protected
{---------------------------------------------------------------}
    procedure Execute;       Override;
    Procedure DoFetchData;
    Procedure DoUpdateIcon;
    Procedure Terminate;

{---------------------------------------------------------------}
  Published
{---------------------------------------------------------------}
    Property Index : Integer Read fIndex
                             Write SetIndex;
    Property MaxIndex : Integer Read FMaxIndex
                                Write SetMaxIndex;
    Constructor Create(Owner : TDirView);
  end;


{===============================================================}
{  TDirView:                                                    }
{===============================================================}


{---------------------------------------------------------------}
  TDirView = class(TCustomListView)
{---------------------------------------------------------------}
  private
{---------------------------------------------------------------}
    {internal properties:}
    FPath                : String;
    FLoadEnabled         : Boolean;
    FDirty               : Boolean;
    FSortColumn          : Integer;
    FSortAscending       : Boolean;
    FConfirmDelete       : Boolean;
    FConfirmOverwrite    : Boolean;
    FShowColumnIcon      : Boolean;
    FDimmHiddenFiles     : Boolean;
    FUseIconCache        : Boolean;
    FFilesSize           : Cardinal;
    FFilesSelSize        : Cardinal;
    FWatchforChanges     : Boolean;
    FLoading             : Boolean;
    FHeaderHandle        : HWND;
    FLastPath            : String;
    FInfoCacheList       : TListExt;
    FDriveView           : TObject;
    FShowDirectories     : Boolean;
    FChangeTimer         : TTimer;
    FChangeInterval      : Cardinal;
    FUseIconUpdateThread : Boolean;
    FIUThreadFinished    : Boolean;
    FReloadTime          : TSystemTime;
    FDirOK               : Boolean;
    FDriveType           : Integer;
    FisRecycleBin        : Boolean;
    FColumnOrder         : TColumnOrder;
    FContextMenu         : Boolean;
    FUseSystemContextMenu: Boolean;
    FParentForm          : TCustomForm;
    FNotifyEnabled       : Boolean;
    FShowMask            : Boolean;
    FShowAnimation       : Boolean;

    {File selection properties:}
    FMask                : String;
    FSelArchive          : TSelAttr;
    FSelHidden           : TSelAttr;
    FSelSysFile          : TSelAttr;
    FSelReadOnly         : TSelAttr;
    FSelFileSizeFrom     : Integer;
    FSelFileSizeTo       : Integer;
    FSelFileDateFrom     : Word;
    FSelFileDateTo       : Word;
    FSelFileTimeFrom     : Word;
    FSelFileTimeTo       : Word;

    {Column properties:}
    FColWidthName        : Integer;
    FColWidthSize        : Integer;
    FColWidthType        : Integer;
    FColWidthDate        : Integer;
    FColWidthAttr        : Integer;

    DateTimeFormatStr    : String;
    FDateTimeDisplay     : TDateTimeDisplay;

    {shFileOperation-shell component TFileOperator:}
    FFileOperator        : TFileOperator;

    {Additional thread components:}
    FIconUpdateThread    : TIconUpdateThread;
    FDiscMonitor         : TDiscMonitor;

    {Additional events:}
    FOnAddFile           : TDirViewAddFileEvent;
    FOnExecFile          : TDirViewExecFileEvent;
    FOnChangeDetected    : TNotifyEvent;
    FOnChangeInvalid     : TNotifyEvent;
    FOnStartLoading      : TNotifyEvent;
    FOnLoaded            : TNotifyEvent;
    FOnSelectItemNew     : TLVSelectItemEvent;

    {Column header bitmaps:}
    ArrowDown, ArrowUp   : HBITMAP;
    FBitmap              : TBitmap;
    FColumnIconPainted   : Boolean;

    iRecycleFolder       : iShellFolder;
    PIDLRecycle          : PItemIDList;

{$IFDEF DRAGDROP}
    {Drag&Drop:}
    FDD                     : TDragDropFilesEx;
    DragFileList            : TStringList;
    FDragDrive              : TDrive;
    FTargetPopUpMenu        : Boolean;
    FUseDragImages          : Boolean;
    FStartPos               : TPoint;
    FDragPos                : TPoint;
    FExeDrag                : Boolean;

    DragStartTime           : FILETIME;
    LastVScrollTime         : FILETIME;
    VScrollCount            : Integer;

    FOnDDDragEnter          : TDDOnDragEnter;
    FOnDDDragLeave          : TDDOnDragLeave;
    FOnDDDragOver           : TDDOnDragOver;
    FOnDDDrop               : TDDOnDrop;
    FOnDDQueryContinueDrag  : TDDOnQueryContinueDrag;
    FOnDDGiveFeedback       : TDDOnGiveFeedback;
    FOnDDDragDetect         : TDDOnDragDetect;
    FOnDDProcessDropped     : TOnProcessDropped;
    FOnDDError              : TDDErrorEvent;
    FOnDDExecuted           : TDDExecutedEvent;
    FOnDDFileOperation      : TDDFileOperationEvent;
    FOnDDFileOperationExecuted : TDDFileOperationExecutedEvent;

    {Internal Drag&Drop eventhandling:}
    Procedure DDDragEnter(DataObj: IDataObject; grfKeyState: Longint; pt: TPoint; var dwEffect: longint; var Accept:boolean);
    Procedure DDDragLeave;
    Procedure DDDragOver(grfKeyState: Longint; pt: TPoint; var dwEffect: longint);
    Procedure DDDrop(DataObj: IDataObject; grfKeyState: Longint;  pt: TPoint; var dwEffect: longint);
    Procedure DDQueryContinueDrag(fEscapePressed: BOOL; grfKeyState: Longint; var Result: HResult);
    Procedure DDGiveFeedback(dwEffect: Longint; var Result: HResult);
    Procedure DDDragDetect(grfKeyState: Longint; DetectStart, Pt: TPoint; DragStatus:TDragDetectStatus);
    Procedure DDProcessDropped(Sender: TObject; grfKeyState: Longint;  pt: TPoint; dwEffect: Longint);

    Procedure SetTargetPopUpMenu(PopMe : Boolean);

    {Drag&drop helper functions:}
    Procedure SignalFileDelete(Sender: TObject; Files : TStringList);
    Procedure PerformDragDropFileOperation(TargetPath : String; dwEffect: Integer);
{$ENDIF}

{---------------------------------------------------------------}
  protected
{---------------------------------------------------------------}
    Procedure Notification(AComponent: TComponent; Operation: TOperation); Override;

    (* Overwritten events: *)
    Procedure ColClick (Column: TListColumn);                     Override;
    Procedure DblClick;                                           Override;
    Procedure Loaded;                                             Override;
    Procedure Change(Item: TListItem; Change: Integer);           Override;
    Procedure KeyDown(var Key: Word; Shift: TShiftState);         Override;
    Procedure CreateWnd;                                          Override;
    Procedure Delete(Item: TListItem);                            Override;
    Procedure Edit(Const HItem: TLVItem);                         Override;
    Function  CanEdit(Item: TListItem): Boolean;                  Override;
{$IFDEF DRAGDROP}
    Procedure DDError(ErrorNo : TDDError);
    Function  GetCanUndoCopyMove : Boolean;                       Virtual;
{$ENDIF}
    Procedure SetColumnImages;                                    Dynamic;
    Procedure SetColumnProperties;                                Dynamic;
    Function  FormatFileTime(FileTime : TFILETIME) : String;      Virtual;
    Function  GetAttrString(Attr : Integer) : String;             Virtual;
    Function  GetItemFromHItem(const Item: TLVItem): TListItem;
    Procedure SetPath(Dir : String);                              Virtual;
    Function  GetPathName : String;
    Procedure SetMask(M : String);                                Virtual;
    Procedure SetWatchForChanges(Watch : Boolean);                Virtual;
    Procedure IconsSetImageList;
    Procedure SetSortAscending(Ascending : Boolean);              Virtual;
    Procedure SetSortColumn(WichCol : TSortColumn);               Virtual;
    Function  GetSortColumn : TSortColumn;
    Procedure SetChangeInterval(Interval : Cardinal);             Virtual;
    Procedure SetShowColumnIcon(ShowIt : Boolean);                Virtual;
    Procedure SetViewStyle(VStyle : TViewStyle);                  Virtual;
    Procedure SetDimmHiddenFiles(Dimm : Boolean);                 Virtual;
    Procedure SetShowDirectories(ShowIt : Boolean);               Virtual;
    Procedure SetDateTimeDisplay(Format: TDateTimeDisplay);       Virtual;
    Procedure LoadFromRecycleBin(Dir : String);                   Virtual;
    Function  GetNextMask(Var Mask : String) : String;            Virtual;
    Procedure SetLoadEnabled(Enabled : Boolean);                  Virtual;
    Procedure CreateColumns;                                      Virtual;
    Procedure SaveColumnProperties;                               Virtual;
    Procedure SetMultiSelect(Select : Boolean);                   Virtual;
    Procedure SetShowMask(Show : Boolean);
    Procedure SetColWidthName(W : Integer);
    Procedure SetColWidthSize(W : Integer);
    Procedure SetColWidthType(W : Integer);
    Procedure SetColWidthDate(W : Integer);
    Procedure SetColWidthAttr(W : Integer);
    Procedure ChangeDetected(Sender: TObject);
    Procedure ChangeInvalid(Sender: TObject);
    Procedure TimerOnTimer(Sender : TObject);
    Procedure UpdateNameHeader(ShowMask : Boolean);

    Procedure InternalOnSelectItem  (Sender: TObject;
                                     Item: TListItem;
                                     Selected: Boolean);

    Procedure InternalOnCustomDrawItem(Sender: TCustomListView;
                                       Item: TListItem;
                                       State: TCustomDrawState;
                                       var DefaultDraw: Boolean);

    Procedure InternalOnCustomDrawSubItem(Sender: TCustomListView;
                                          Item: TListItem;
                                          SubItem: Integer;
                                          State: TCustomDrawState;
                                          var DefaultDraw: Boolean);

    Procedure SetDateTimeFormatString;                            Dynamic;

    Function  GetColumnOrder : TColumnOrder;                      Virtual;
    Procedure SetColumnOrder(Order: TColumnOrder);                Virtual;

    {Hooked windows messages:}
    Procedure CNNotify(Var Msg: TWMNotify);                Message CN_NOTIFY;
    Procedure WMNotify(Var msg: TWMNotify);                Message WM_NOTIFY;
    Procedure WMPaint(Var Msg : TWMPaint);                 Message WM_PAINT;
{$IFDEF DRAGDROP}
    Procedure WMLButtonDown(Var Msg: TWMLButtonDown);      Message WM_LBUTTONDOWN;
{$ENDIF}
    Procedure WMRButtonDown(Var Msg: TWMRButtonDown);      Message WM_RBUTTONDOWN;
    Procedure WMRButtonUp(Var Msg: TWMRButtonUp);          Message WM_RBUTTONUP;
    Procedure WMDestroy(Var Msg: TWMDestroy);              Message WM_DESTROY;

{---------------------------------------------------------------}
  public
{---------------------------------------------------------------}
    {Runtime, readonly properties:}
    Property FilesSize      : Cardinal    Read fFilesSize;
    Property FilesSelSize   : Cardinal    Read fFilesSelSize;
    Property LastPath       : String      Read fLastPath;
    Property ReloadTime     : TSystemTime Read fReloadTime;
    Property DirOK          : Boolean     Read fDirOK;
    Property DriveType      : Integer     Read fDriveType;
    Property isRecycleBin   : Boolean     Read fisRecycleBin;

    {Runtime properies:}
    {Enable or disable populating the item list:}
    Property LoadEnabled    : Boolean     Read FLoadEnabled
                                          Write SetLoadEnabled
                                          Default True;

    {Displayed data is not valid => reload required}
    Property Dirty          : Boolean     Read FDirty;

    {Populating of items is in progress:}
    Property Loading        : Boolean     Read FLoading
                                          Write FLoading;

    {Returns/requires the actual directory without trailing slash: C:, C:\WINDOWS }
    Property Path : String Read fPath
                           Write SetPath;


    {Returns the actual directory with trailing slash, when root-directory: C:\, C:\WINDOWS }
    Property PathName   : String  Read GetPathName;

    {Linked component TDriveView:}
    Property DriveView  : TObject Read fDriveView
                                  Write fDriveView;

    {TImagelists for the system image list: !!do not change!!}
    Property SmallImages;
    Property LargeImages;

    {It is not required to store the items edited at designtime:}
    Property Items Stored False;

    {Column order:}
    Property ColumnOrder : TColumnOrder Read GetColumnOrder
                                        Write SetColumnOrder;

{$IFDEF DRAGDROP}
    {Drag&Drop runtime, readonly properties:}
    Property DragDropFilesEx : TDragDropFilesEx Read FDD;
    Property DragDrive       : TDrive Read FDragDrive;
    Property CanUndoCopyMove : Boolean Read GetCanUndoCopyMove;
    Property DDFileOperator  : TFileOperator Read FFileOperator;

    {Drag&Drop fileoperation methods:}
    Function  UndoCopyMove : Boolean;                          Dynamic;

    {Clipboard fileoperation methods (requires drag&drop enabled):}
    Procedure EmptyClipboard;                                  Dynamic;
    Function  CopyToClipBoard : Boolean;                       Dynamic;
    Function  CutToClipBoard : Boolean;                        Dynamic;
    Function  CanPasteFromClipBoard : Boolean;                 Dynamic;
    Function  PasteFromClipBoard : Boolean;                    Dynamic;
{$ENDIF}

    {Thread handling: }
    procedure StartWatchThread;
    procedure StopWatchThread;
    Function  WatchThreadActive : Boolean;
    Procedure StartIconUpdateThread;
    Procedure StopIconUpdateThread;
    Procedure TerminateThreads;
    Procedure Syncronize;


    {Other additional functions: }
    Procedure ClearIconCache;
    Procedure BuildFileList(List : TStringList);                   Dynamic;
    Function  NormalizeMask(Mask : String) : String;               Dynamic;

    {Create a new file:}
    Function  CreateFile(NewName : String) : TListItem;            Dynamic;
    {Delete all selected files:}
    Function  DeleteSelectedFiles(AllowUndo : Boolean) : Boolean;  Dynamic;
    {Execute specified file:}
    Procedure ExecFile(Item : TListItem);                          Dynamic;

    {Check, if file or files still exists:}
    Procedure ValidateFile(Item     : TListItem);                  OverLoad;
    Procedure ValidateFile(FileName :TFileName);                   OverLoad;
    Procedure ValidateSelectedFiles;                               Dynamic;

    {Access the internal data-structures:}
    Function  AddItem(Srec : TSearchRec) : TListItem;
    Procedure GetDisplayData(Item : TListItem; FetchIcon : Boolean);
    function  GetFileRec(Index: Integer): PFileRec;

    {Populate / repopulate the filelist:}
    procedure Load(Dir : String);
    Procedure ReLoad(CacheIcons : Boolean);

    {Set or unset selection mark for files:}
    Procedure SelectFiles(Filter : TFileFilter; Select : Boolean);

    {Filename can be accessed by these functions:}
    Function  GetFullFileName(Item : TListItem) : String;
    Function  GetFileName(Item : TListItem)     : String;
    Function  GetFileNameOnly(Item : TListItem) : String;
    Function  GetFileExt(Item : TListItem)      : String;

    {Menu-handling:}
    {Diplays the shell's properties sheet for all selected files:}
    Procedure DisplayPropertiesMenu;                               Dynamic;
    {Diplays the shell's context menu  for all selected files:}
    Procedure DisplayContextMenu;                                  Dynamic;
    Function  FindFileItem(FileName : TFileName) : TListItem;      Dynamic;

    Constructor Create(AOwner: TComponent);                        Override;
    Destructor  Destroy;                                           Override;

    {Redefined functions: }
    Function  CustomSort(SortProc: TLVCompare; Data: Longint): Boolean; virtual;
    Function  AlphaSort: Boolean;


{---------------------------------------------------------------}
  published
{---------------------------------------------------------------}
    {Additional properties:}

    {Set the sort column of the listview}
    Property SortColumn : TSortColumn Read  GetSortColumn
                                      Write SetSortColumn
                                      Default sName;


    {Sortorder of actual sort column}
    Property SortAscending : Boolean Read fSortAscending
                                     Write SetSortAscending
                                     Default True;

    {Column width properties:}
    Property ColWidthName : Integer Read FColWidthName
                                    Write SetColWidthName
                                    Default 150;
    Property ColWidthSize : Integer Read FColWidthSize
                                    Write SetColWidthSize
                                    Default 80;
    Property ColWidthType : Integer Read FColWidthType
                                    Write SetColWidthType
                                    Default 125;
    Property ColWidthDate : Integer Read FColWidthDate
                                    Write SetColWidthDate
                                    Default 130;
    Property ColWidthAttr : Integer Read FColWidthAttr
                                    Write SetColWidthAttr
                                    Default 45;

    {Confirm deleting files}
    Property ConfirmDelete       : Boolean Read fConfirmDelete
                                           Write fConfirmDelete
                                           Default True;

    {Confirm overwriting files}
    Property ConfirmOverwrite    : Boolean Read fConfirmOverwrite
                                           Write fConfirmOverwrite
                                           Default True;

    {Watch current directory for filename changes (create, rename, delete files)}
    Property WatchForChanges : Boolean  Read fWatchForChanges
                                        Write SetWatchForChanges;

    {Reload the directory after only the interval:}
    Property ChangeInterval  : Cardinal Read fChangeInterval
                                        Write SetChangeInterval
                                        Default 1000;

    {Fetch shell icons by thread:}
    Property UseIconUpdateThread : Boolean Read  fUseIconUpdateThread
                                           Write fUseIconUpdateThread;


    {Show or hide the sort symbol in the column header:}
    Property ShowColumnIcon      : Boolean Read fShowColumnIcon
                                           Write SetShowColumnIcon
                                           Default True;

    {Paint hidden files in gray color:}
    Property DimmHiddenFiles     : Boolean Read fDimmHiddenFiles
                                           Write SetDimmHiddenFiles;

    {Show subdirectories in the listview:}
    Property ShowDirectories     : Boolean Read fShowDirectories
                                           Write SetShowDirectories
                                           Default True;

    {Enables or disables the system context menu for a file or directory:}
    Property UseSystemContextMenu : Boolean Read FUseSystemContextMenu
                                           Write FUseSystemContextMenu
                                           Default True;

    {Enables or disables icon caching for registered file extensions. Caching enabled
     enhances the performance but does not take care about installed icon handlers, wich
     may modify the display icon for registered files. Only the iconindex is cached not the
     icon itself:}
    Property UseIconCache        : Boolean Read FUseIconCache
                                           Write FUseIconCache;

    {Display format of the date/time of the files:}
    Property DateTimeDisplay     : TDateTimeDisplay Read fDateTimeDisplay
                                                    Write SetDateTimeDisplay;

    {Properties for filtering files:}

    {filemask, multiple filters are possible: '*.pas;*.dfm'}
    Property Mask        : String          Read fMask
                                           Write SetMask;

    {File-attributes selection:}
    Property SelArchive  : TSelAttr         Read    fSelArchive
                                            Write   fSelArchive
                                            Default selDontCare;

    Property SelHidden   : TSelAttr         Read    fSelHidden
                                            Write   fSelHidden
                                            Default selDontCare;

    Property SelSysFile  : TSelAttr         Read    fSelSysFile
                                            Write   fSelSysFile
                                            Default selDontCare;

    Property SelReadOnly : TSelAttr         Read    fSelReadOnly
                                            Write   fSelReadOnly
                                            Default selDontCare;

    {Filesize selection:}
    Property SelFileSizeFrom : Integer      Read    fSelFileSizeFrom
                                            Write   fSelFileSizeFrom;

    Property SelFileSizeTo   : Integer      Read    fSelFileSizeTo
                                            Write   fSelFileSizeTo
                                            Default MaxInt;

    {File date selection: 01.01.1980 to 31.12.2099}
    {See functions DosDateToDateStr and StrDateToDosDate for conversion.}
    Property SelFileDateFrom : Word         Read    fSelFileDateFrom
                                            Write   fSelFileDateFrom
                                            Default MinDate; {01.01.1980}

    Property SelFileDateTo   : Word         Read    fSelFileDateTo
                                            Write   fSelFileDateTo
                                            Default MaxDate; {31.12.2099}

    {File time selection: 00:00:00 to 24:00:00}
    Property SelFileTimeFrom : Word         Read    fSelFileTimeFrom
                                            Write   fSelFileTimeFrom;

    Property SelFileTimeTo   : Word         Read    fSelFileTimeTo
                                            Write   fSelFileTimeTo
                                            Default MaxTime;

    Property ShowMaskInHeader : Boolean     Read    FShowMask
                                            Write   SetShowMask;

    Property ShowAnimation    : Boolean     Read    FShowAnimation
                                            Write   FShowAnimation;

    {Additional events:}

    {The watchthread has detected new, renamed or deleted files}
    Property OnChangeDetected : TNotifyEvent Read fOnChangeDetected
                                            Write fOnChangeDetected;


    {The watchthread can't watch the current directory. Occurs on novell
     network drives.}
    Property OnChangeInvalid  : TNotifyEvent Read fOnChangeInvalid
                                            Write fOnChangeInvalid;

    {This event is fired, when The listview is completly populated}
    Property OnLoaded         : TNotifyEvent Read fOnLoaded
                                            Write fOnLoaded;

    {This event is fired, before populating the listview}
    Property OnStartLoading   : TNotifyEvent Read fOnStartLoading
                                            Write fOnStartLoading;

    {Set AddFile to false, if actual file should not be added to the filelist:}
    Property OnAddFile        : TDirViewAddFileEvent Read fOnAddFile
                                                     Write fOnAddFile;

    {Set AllowExec to false, if actual file should not be executed:}
    Property OnExecFile      : TDirViewExecFileEvent Read fOnExecFile
                                                     Write fOnExecFile;

    {Redefined event OnSelectItemNew. The original event ist used for
     internal purposes:}
    Property OnSelectItemNew  : TLVSelectItemEvent Read fOnSelectItemNew
                                                   Write fOnSelectItemNew;

{$IFDEF DRAGDROP}
    {Drag&Drop:}

    {Show drag images during a drag&drop operation:}
    Property UseDragImages         : Boolean      Read FUseDragImages
                                                  Write FUseDragImages
                                                  Default True;

    {Show popupmenu when dropping a file with the right mouse button:}
    Property TargetPopUpMenu       : Boolean      Read FTargetPopUpMenu
                                                  Write SetTargetPopUpMenu
                                                  Default True;

    {The mouse has entered the component window as a target of a drag&drop operation:}
    Property OnDDDragEnter         : TDDOnDragEnter Read FOnDDDragEnter
                                                    Write FOnDDDragEnter;

    {The mouse has leaved the component window as a target of a drag&drop operation:}
    Property OnDDDragLeave         : TDDOnDragLeave Read FOnDDDragLeave
                                                    Write FOnDDDragLeave;

    {The mouse is dragging in the component window as a target of a drag&drop operation:}
    Property OnDDDragOver          : TDDOnDragOver  Read FOnDDDragOver
                                                    Write FOnDDDragOver;
    {The Drag&drop operation is about to be executed:}
    Property OnDDDrop              : TDDOnDrop      Read FOnDDDrop
                                                    Write FOnDDDrop;

    Property OnDDQueryContinueDrag : TDDOnQueryContinueDrag  Read FOnDDQueryContinueDrag
                                                             Write FOnDDQueryContinueDrag;

    Property OnDDGiveFeedback      : TDDOnGiveFeedback Read FOnDDGiveFeedback
                                                       Write FOnDDGiveFeedback;

    {A drag&drop operation is about to be initiated whith the components window as the
     source:}
    Property OnDDDragDetect        : TDDOnDragDetect   Read FOnDDDragDetect
                                                       Write FOnDDDragDetect;

    {The component window is the target of a drag&drop operation:}
    Property OnDDProcessDropped    : TOnProcessDropped Read FOnDDProcessDropped
                                                       Write FOnDDProcessDropped;

    {An error has occured during a drag&drop operation:}
    Property OnDDError             : TDDErrorEvent Read FOnDDError
                                                   Write FOnDDError;

    {The drag&drop operation has been executed:}
    Property OnDDExecuted          : TDDExecutedEvent Read FOnDDExecuted
                                                      Write FOnDDExecuted;
    {Event is fired just before executing the fileoperation. This event is also fired when
     files are pasted from the clipboard:}
    Property OnDDFileOperation     : TDDFileOperationEvent Read FOnDDFileOperation
                                                           Write FOnDDFileOperation;

    {Event is fired after executing the fileoperation. This event is also fired when
     files are pasted from the clipboard:}
    Property OnDDFileOperationExecuted  : TDDFileOperationExecutedEvent Read FOnDDFileOperationExecuted
                                                                        Write FOnDDFileOperationExecuted;
{$ENDIF}
    property Align;
    property AllocBy;
    property Anchors;
    property BiDiMode;
    property BorderStyle;
    property BorderWidth;
    property Checkboxes;
    property Color;
    property Columns Stored False;  {<== required, otherwise AV generated, when dragging columns}
    property ColumnClick Default True;
    property Constraints;
    property Ctl3D;
    property Enabled;
    property Font;
    property FlatScrollBars;
    property FullDrag;
    property GridLines;
    property HideSelection;
    property HotTrack;
    property HotTrackStyles;
    property IconOptions;
    property MultiSelect Write SetMultiSelect;
    property ReadOnly default False;
    Property RowSelect;
    property ParentBiDiMode;
    property ParentColor default False;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowColumnHeaders;
    property ShowHint;
    property TabOrder;
    property TabStop default True;
    property ViewStyle Write SetViewStyle;
    property Visible;
    property OnChange;
    property OnChanging;
    property OnClick;
    property OnColumnClick;
    property OnCustomDraw;
    {Used for internal purposes:
    property OwnerDraw;
    property OnCustomDrawItem;
    property OnCustomDrawSubItem;}
    {Not compatible properties with this implementation:
    property OwnerData;
    property OnGetImageIndex;
    property OnCompare;
    property OnData;
    property OnDataFind;
    property OnDataHint;
    property OnDataStateChange;}
    property OnDblClick;
    property OnDeletion;
    property OnDrawItem;
    property OnEdited;
    property OnEditing;
    property OnEndDock;
    property OnEnter;
    property OnExit;
    property OnInsert;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDock;
  end; {Type TDirView}


procedure Register;


{Additional non-component specific functions:}

{Create and resolve a shell link (file shortcut):}
Function  CreateFileShortCut(SourceFile, Target, DisplayName : String) : Boolean;
Function  ResolveFileShortCut(SourceFile : String) : String;

{Sets bitmap bordercolor to systemcolor for 3D-Objects:}
Procedure FixBitmap(hbmp: HBITMAP);

{Gets the shell's display icon for registered file extensions:}
Function GetIconIndex(Const AFile: String; Attrs: DWORD; Flags: UINT): Integer;

{Gets the shell's inforecord for registered fileextensions:}
Function GetshFileInfo(Const AFile: String; Attrs: DWORD; Flags: UINT): TSHFileInfo;

{Returns the displayname as used by the shell:}
Function GetShellDisplayName(Const ShellFolder: IShellFolder; IDList: PItemIDList; Flags: DWORD; var Name: string): boolean;

{Returns True, if the specified file is a executable file:}
Function isExecutable(Filename: String): Boolean;

{Returns True, if the specified extension matches one of the extensions in ExtList:}
Function MatchesFileExt(Ext : ExtStr; Const FileExtList : String) : Boolean;

{Returs, wether COMCTL32 supports the extended display properties:
 COMCTL32.DLL version 4.70 or higher ist required. Version 4.70 is
 included in Internet Explorer 4 with Active Desktop.
 Updates of COMCTL32.DLL are available at:
 http://msdn.microsoft.com/developer/downloads/files/40Comupd.htm }
Function COMCTL32OK : Boolean;

Var HasExtendedCOMCTL32 : Boolean;
    UnKnownFileIcon     : Integer;
    DefaultExeIcon      : Integer;
    StdDirTypeName      : String;
    StdDirIcon          : Integer;
    StdDirSelIcon       : Integer;
    LastClipBoardOperation : TClipBoardOperation;
    DropSource          : TObject;
    DragImageList       : TDragImageList;
    WinDir              : String;
    TempDir             : String;
    ErrorOpenFile       : String;
    ErrorRenameFile     : String;
    ErrorInvalidName    : String;
    TextFileExt         : String;
    TextFiles           : String;
    TextDirectories     : String;


{===============================================================}
implementation
{===============================================================}

Uses DriveView;


ResourceString
  {German messages:}
  GERMAN_ColIDName        = 'Name';
  GERMAN_ColIDSize        = 'Gre';
  GERMAN_ColIDType        = 'Typ';
  GERMAN_ColIDAttr        = 'Attr';
  GERMAN_ColIDDate        = 'Gendert am';
  GERMAN_ErrorOpenFile    = 'Fehler ffnen der Datei: ';
  GERMAN_ErrorRenameFile  = 'Datei oder Verzeichnis kann nicht umbenannt werden:';
  GERMAN_ErrorInvalidName = 'Dateiname enthlt ungltige Zeichen:';
  GERMAN_TextFileExt      = 'Datei %s';
  GERMAN_TextFiles        = '%u Dateien';
  GERMAN_TextDirectories  = '%u Verzeichnisse';

  {English messages:}
  ENGLISH_ColIDName       = 'Name';
  ENGLISH_ColIDSize       = 'Size';
  ENGLISH_ColIDType       = 'Type';
  ENGLISH_ColIDAttr       = 'Attr';
  ENGLISH_ColIDDate       = 'Changed';
  ENGLISH_ErrorOpenFile   = 'Can''t open file: ';
  ENGLISH_ErrorRenameFile = 'Can''t rename file or directory: ';
  ENGLISH_ErrorInvalidName= 'Filename contains invalid characters:';
  ENGLISH_TextFileExt     = 'File %s';
  ENGLISH_TextFiles       = '%u Files';
  ENGLISH_TextDirectories = '%u Directories';

  {French messages by dsnake@infonie.fr:}
  FRENCH_ColIDName        = 'Nom';
  FRENCH_ColIDSize        = 'Taille';
  FRENCH_ColIDType        = 'Type';
  FRENCH_ColIDAttr        = 'Attributs';
  FRENCH_ColIDDate        = 'Date';
  FRENCH_ErrorOpenFile    = 'Ne peux ouvrir le fichier: ';
  FRENCH_ErrorRenameFile  = 'Ne peux renommer le fichier: ';
  FRENCH_ErrorInvalidName = 'Le nom de fichier contient des caractres invalides:';
  FRENCH_TextFileExt      = 'Fichier %s';


procedure Register;
begin
  RegisterComponents('IE', [TDirView]);
end; {Register}



function CompareInfoCacheItems (I1, I2: Pointer): Integer;
Begin
  IF PInfoCache(I1)^.FileExt < PInfoCache(I2)^.FileExt Then Result := fLess    Else
  IF PInfoCache(I1)^.FileExt > PInfoCache(I2)^.FileExt Then Result := fGreater Else
                                                    Result := fEqual;
End; {CompareInfoCache}


function CompareIconCacheItems (I1, I2: Pointer): Integer;
Begin
  IF PIconRec(I1)^.FileName < PIconRec(I2)^.FileName Then Result := fLess    Else
  IF PIconRec(I1)^.FileName > PIconRec(I2)^.FileName Then Result := fGreater Else
                                                    Result := fEqual;
End; {CompareIconCacheItems}


{=============================================================}
{                     TIconUpdateThread:                      }
{=============================================================}




Constructor TIconUpdateThread.Create(Owner : TDirView);
Begin
  Inherited Create(True);
  fOwner := Owner;
  fIndex := 0;
  fNewIcons := False;
  IF (FOwner.ViewStyle = vsReport) Or (FOwner.ViewStyle = vsList) Then
  fMaxIndex := fOwner.VisibleRowCount
  Else
  FMaxIndex := 0;
  fOwner.fIUThreadFinished := False;
end;  {TIconUpdateThread.Create}


Procedure TIconUpdateThread.SetMaxIndex(i : Integer);
Var P    : TPoint;
    Item : TListItem;

Begin
  IF i <> FMaxIndex Then
  Begin
    fNewIcons := True;
    IF i < fMaxIndex Then
    Begin
      IF Suspended Then
      FIndex := i
      Else
      Begin
        P.x := 0;
        P.y := 0;
        Item := FOwner.GetNearestItem(P, TSearchDirection(sdAbove));
        IF Assigned(Item) Then
          FIndex := Item.Index
        Else
        FIndex := i;
      End;
    End
    Else
    FMaxIndex := i;
  End;
End; {SetMaxIndex}


Procedure TIconUpdateThread.SetIndex(i : Integer);
Var PageSize : Integer;

Begin
  If i <> fIndex Then
  Begin
    PageSize := fOwner.VisibleRowCount;
    fIndex := i;
    fNewIcons := True;
    IF FOwner.ViewStyle = VsList Then
    fMaxIndex := i + 2 * PageSize
    Else
    fMaxIndex := i + PageSize;
  End;
End; {SetIndex}


procedure TIconUpdateThread.Execute;

Var FileInfo     : TShFileInfo;
    Count         : Integer;

Begin
  IF Assigned(fOwner.TopItem) Then
  fIndex := fOwner.TopItem.Index
  Else
  fIndex := 0;

  fNewIcons := fIndex > 0;

  While Not Terminated Do
  Begin
    IF (fIndex  > fMaxIndex) Then
    Suspend;
    Count := fOwner.Items.Count;
    IF Not Terminated And ((fIndex >= Count) Or (Count = 0)) Then
      Suspend;
    InvalidItem := True;
    IF Terminated Then
    Break;
    Synchronize(DoFetchData);
    IF Not InvalidItem And
       Not Terminated  And
      CurrentItemData.IconEmpty And
      (Pos(CurrentItemData.FileExt, SpecialExtensions) <> 0) Then
    Begin
      Try
        shGetFileInfo(PChar(CurrentFilePath), FILE_ATTRIBUTE_NORMAL, FileInfo, SizeOf(FileInfo),
                      SHGFI_TYPENAME Or SHGFI_USEFILEATTRIBUTES Or SHGFI_SYSICONINDEX Or SHGFI_ICON);
      Except Else fSyncIcon := UnKnownFileIcon; End; {Capture exceptions generated by the shell}
      IF Terminated Then
      Break;
      fSyncIcon := FileInfo.iIcon;
      IF fSyncIcon <> CurrentItemData.ImageIndex Then
      fNewIcons := True;
      IF Not Terminated Then
        Synchronize(DoUpdateIcon);
    End;
    SetLength(CurrentFilePath, 0);
    IF CurrentIndex = fIndex Then
    Inc(fIndex);
    SetLength(CurrentFilePath, 0);
  End;
End; {TIconUpdateThread.Execute}


Procedure TIconUpdateThread.DoFetchData;
Begin
  CurrentIndex := fIndex;
  IF Not Terminated                             And
     (Pred(fOwner.Items.Count) >= CurrentIndex) And
     Assigned(fOwner.Items[CurrentIndex])       And
     Assigned(fOwner.Items[CurrentIndex].Data)  Then
  Begin
    CurrentFilePath := fOwner.GetFullFileName(fOwner.Items[CurrentIndex]);
    CurrentItemData := PFileRec(fOwner.Items[CurrentIndex].Data)^;
    InvalidItem     := False;
  End
  Else
  InvalidItem := True;
End; {TIconUpdateThread.DoFetchData}


Procedure TIconUpdateThread.DoUpdateIcon;
Var LVI : TLVItem;
Begin
  IF (fOwner.Items.Count > CurrentIndex) And
  Not fOwner.Loading And
  Not Terminated And
  {(CurrentIndex = fIndex) And}
  Assigned(fOwner.Items[CurrentIndex]) And
  Assigned(fOwner.Items[CurrentIndex].Data) Then
  Begin
    With fOwner.Items[CurrentIndex] Do
    Begin
      IF (fSyncIcon >= 0) And (PFileRec(Data)^.ImageIndex <> fSyncIcon) Then
      Begin
        PFileRec(Data)^.ImageIndex := fSyncIcon;

        {To avoid flickering of the display use Listview_SetItem
        instead of tht property ImageIndex:}
        LVI.mask := LVIF_IMAGE;
        LVI.iItem := CurrentIndex;
        LVI.iSubItem := 0;
        LVI.iImage := fSyncIcon;
        IF Not Terminated Then
        ListView_SetItem(fOwner.Handle, LVI);
        fNewIcons := True;
      End;
      PFileRec(Data)^.IconEmpty := False;
    End;
  End;
End; {TIconUpdateThread.DoUpdateIcon}


Procedure TIconUpdateThread.Terminate;
Begin
  fOwner.fIUThreadFinished := True;
  Inherited Terminate;
End; {TIconUpdateThread.Terminate}


{=============================================================}
{                           TDirView                          }
{=============================================================}

constructor TDirView.Create(AOwner: TComponent);
Var WinVer : TOSVersionInfo;

Begin
  Inherited Create(AOwner);
  WinVer.dwOSVersionInfoSize := SizeOf(WinVer);
  GetVersionEx(WinVer);
  fInfoCacheList   := TListExt.Create(SizeOf(TInfoCache));
  fSortColumn      := 0;
  fSortAscending   := True;
  FFilesSize       := 0;
  FFilesSelSize    := 0;
  fLoading         := False;
  FLoadEnabled     := True;
  FDirty           := False;
  fMask            := '*.*';
  fLastPath        := '';
  fShowColumnIcon  := True;
  fDirOK           := False;
  fDriveType       := DRIVE_UNKNOWN;
  fIsRecycleBin    := False;
  FUseIconCache    := False;
  fConfirmDelete   := True;
  FContextMenu     := False;
  FNotifyEnabled   := True;
  FShowDirectories := True;
  FUseSystemContextMenu := True;
  FColumnOrder[0]  := 0;
  FColumnOrder[1]  := 1;
  FColumnOrder[2]  := 2;
  FColumnOrder[3]  := 3;
  FColumnOrder[4]  := 4;
  FColumnIconPainted := False;

  {Redirected events:}
  OnSelectItem        := InternalOnSelectItem;
  OnCustomDrawItem    := InternalOnCustomDrawItem;
  OnCustomDrawSubItem := InternalOnCustomDrawSubItem;


  FFileOperator := TFileOperator.Create(Self);
  FFileOperator.ProgressTitle := coFileOperatorTitle;
  FFileOperator.Flags := [foAllowUndo, foNoConfirmMkDir];

  FColWidthName := 150;
  FColWidthSize := 80;
  FColWidthType := 125;
  FColWidthDate := 130;
  FColWidthAttr := 45;

  CreateColumns;

  FDiscMonitor := NIL;

  {Load Column-Bitmaps:}
  ArrowUp:=LoadBitMap(hInstance, 'UPARROW2');
  ArrowDown:=LoadBitmap(hInstance, 'DOWNARROW2');
  FixBitMap(ArrowUp);
  FixBitMap(ArrowDown);
  FBitmap:=TBitmap.Create;


  {ChangeTimer: }
  IF fChangeInterval = 0 Then
  fChangeInterval       := 1000;
  fChangeTimer := TTimer.Create(Self);
  fChangeTimer.Interval := fChangeInterval;
  fChangeTimer.Enabled  := False;
  fChangeTimer.OnTimer  := TimerOnTimer;

  fSelArchive    := selDontCare;
  fSelHidden     := selDontCare;
  fSelReadOnly   := selDontCare;
  fSelSysFile    := selDontCare;
  fSelFileSizeTo := High(fSelFileSizeTo);
  fSelFileDateFrom := $21; {01.01.1980}
  fSelFileDateTo := $EF9F; {31.12.2099}
  fSelFileTimeTo := $C000; {24:00:00}


{$IFDEF DRAGDROP}
  {Drag&drop:}
  fConfirmOverwrite := True;
  FTargetPopupMenu  := True;
  FDragDrive        := #0;
  FUseDragImages    := (Win32PlatForm = VER_PLATFORM_WIN32_NT) Or (WinVer.dwMinorVersion > 0);
  FStartPos.X       := -1;
  FStartPos.Y       := -1;
  FDragPos          := FStartPos;
  FExeDrag          := False;
  FDD := TDragDropFilesEx.Create(Self);
  With FDD Do
  Begin
    AutoDetectDnD    := False;
    AcceptOwnDnD     := True;
    BringToFront     := True;
    CompleteFileList := True;
    DragDetectDelta  := 2;
    NeedValid        := [nvFileName];
    RenderDataOn     := rdoEnterAndDropSync;
    SourceEffects    := [deCopy, deMove, deLink];
    TargetEffects    := [deCopy, deMove, deLink];
    TargetPopUpMenu  := FTargetPopupMenu;
    OnDragDetect     := DDDragDetect;
    OnDragEnter      := DDDragEnter;
    OnDragLeave      := DDDragLeave;
    OnDragOver       := DDDragOver;
    OnProcessDropped := DDProcessDropped;
    OnDrop           := DDDrop;
    OnQueryContinueDrag := DDQueryContinueDrag;
    OnGiveFeedback   := DDGiveFeedback;
  End;
  DragFileList := TStringList.Create;
{$ENDIF}
  SetDateTimeFormatString;
End; {Create}


Procedure TDirView.WMDestroy(Var Msg: TWMDestroy);
Begin
  TerminateThreads;
  Inherited;
End; {WMDestroy}


Destructor TDirView.Destroy;
Begin
{$IFDEF DRAGDROP}
  DragFileList.Free;
  IF Assigned(FDD) Then
  FDD.Free;
{$ENDIF}

  IF Assigned(PIDLRecycle) Then
  PIDL_Free(PIDLRecycle);

  fInfoCacheList.Free;

  IF Assigned(DragImageList) Then
  DragImageList.Free;
  DragImageList := NIL;

  IF Assigned(SmallImages) Then
  SmallImages.Free;

  IF Assigned(LargeImages) Then
  LargeImages.Free;

  IF Assigned(FFileOperator) Then
  FFileOperator.Destroy;
  FBitmap.ReleaseHandle;
  FBitmap.Free;

  IF Assigned(fChangeTimer) Then
  fChangeTimer.Free;

  DeleteObject(ArrowUp);
  DeleteObject(ArrowDown);
  Inherited Destroy;
End; {Destroy}


Procedure TDirView.TerminateThreads;
Begin
  StopIconUpdateThread;
  if Assigned(FDiscMonitor) then
  Begin
    {FDiscMonitor.Open;}
    FDiscMonitor.Destroy;
    FDiscMonitor := NIL;
  End;
End; {TerminateThreads}


Procedure TDirView.SetPath(Dir : String);
Begin
 While (Length(Dir) > 0) And (Dir[Length(Dir)] = '\') Do
 SetLength(Dir, Pred(Length(Dir)));
 IF fPath <> Dir Then
 Begin
   fLastPath := fPath;
   fPath := Dir;
   Load(fPath);
 End;
 LoadEnabled := True;
End; {SetPath}


Procedure TDirView.SetLoadEnabled(Enabled : Boolean);
Begin
  IF Enabled <> LoadEnabled Then
  Begin
    FLoadEnabled := Enabled;
    IF Enabled And FDirty Then
    Reload(True);
  End;
End; {SetLoadEnabled}


Procedure TDirView.SetMultiSelect(Select : Boolean);
Begin
  If Select <> MultiSelect Then
  Begin
    SaveColumnProperties;
    Inherited MultiSelect := Select;
    CreateColumns;
    IF DirOK Then
    Reload(True);
  End;
End; {SetMultiSelect}


Function TdirView.GetPathName : String;
Begin
 IF (Length(fPath) = 2) And (fPath[2] = ':') Then
 Result := fPath + '\'
 Else
 Result := fPath;
End; {GetPathName}


Procedure TDirView.IconsSetImageList;
var FileInfo        : TShFileInfo;

Begin
  {Get/set system-imagelist for small icons:}
  IF Not Assigned(SmallImages) Then
  Begin
    SmallImages  := TImageList.Create(Self);
    SmallImages.Handle       := SHGetFileInfo('', 0, FileInfo, SizeOf(FileInfo), SHGFI_SYSICONINDEX or SHGFI_SMALLICON);
    SmallImages.ShareImages  := True;
  End;

  {Get/set system-imagelist for large icons:}
  IF Not Assigned(LargeImages) Then
  Begin
    LargeImages := TImageList.Create(Self);
    LargeImages.Handle := SHGetFileInfo('', 0, FileInfo, SizeOf(FileInfo), SHGFI_SYSICONINDEX or SHGFI_LARGEICON);
    LargeImages.ShareImages := True;
  End;
End; {IconsSetImageList}


Procedure TDirView.ExecFile(Item : TListItem);
Var retCode : HINST;
    FileName : String;
Begin
IF Assigned(Item) Then
  Begin
  FileName := GetFullFileName(Item);
  RetCode := ShellExecute(Application.Handle,
                          {'open',} NIL,
                          PChar(FileName),
                          '',
                          PChar(PathName),
                          SW_SHOWNORMAL);
    IF RetCode <= 32 Then
      Raise EExecFile.Create(ErrorOpenFile + FileName);
  End;
End; {ExecFile}


function TDirView.GetFileRec(Index: Integer): PFileRec;
begin
  IF Index > Pred(Items.Count) Then
  Result := NIL
  Else
  Result := Items[index].Data;
end; {GetFileRec}


Function TDirView.AddItem(Srec : TSearchRec) : TListItem;
Var PItem     : PFileRec;
    Item      : TListItem;
Begin
  Item := Items.Add;
  Item.Caption := SRec.Name;
  New(PItem);
  With PItem^ Do
  Begin
    FileName    := Srec.Name;
    DisplayName := Srec.Name;
    Attr        := Srec.FindData.dwFileAttributes;
    Size        := Srec.Size;
    FileTime    := Srec.FindData.ftLastWriteTime;
    Empty       := True;
    isDirectory := Attr and faDirectory <> 0;
    isRecycleBin:= isDirectory               And
                   (Length(Path) = 2)        And
                   (Attr And faSysFile <> 0) And
                   (UpperCase(FileName) = 'RECYCLED');
    IconEmpty   := True;
    INC(FFilesSize, Size);
    FileExt     := UpperCase(Copy(ExtractFileExt(Srec.Name), 2, 3));
    PIDL        := NIL;

    Item.Data    := PItem;

    {Add empty subitems:}
    {Item.SubItems.CommaText := ',,,,';}
    With Item.SubItems Do
    Begin
      Add(EmptyStr);
      Add(EmptyStr);
      Add(EmptyStr);
      Add(EmptyStr);
    End;

    IF FileExt = 'LNK' Then
    Item.OverlayIndex := 1;

  end;
  Result := Item;
End; {AddItem}


Procedure TDirView.LoadFromRecycleBin(Dir : String);
Var iDesktop         : iShellFolder;
    PIDLRecycleLocal : PItemIDList;
    PCurrList        : PItemIDList;
    FQPIDL           : PItemIDList;
    EnumList         : IEnumIDList;
    Fetched          : ULONG;
    SRec             : TSearchRec;
    DisplayName      : String;
    FullPath         : String;
    NewItem          : TListItem;
    FileRec          : PFileRec;
    FileInfo         : TSHFileInfo;
    FileSel          : Boolean;
    ItemsAdded       : Integer;
    StartTime        : DWORD;
    Animate          : TAnimate;


Begin
  If Not Assigned(iRecycleFolder) Then
  Begin
    OLECheck(shGetDesktopFolder(iDesktop));
    OLECheck(shGetSpecialFolderLocation(Self.Handle, CSIDL_BITBUCKET, PIDLRecycleLocal));
    PIDLRecycle := PIDL_Concatenate(NIL, PIDLRecycleLocal);
    If Not SUCCEEDED(iDesktop.BindToObject(PIDLRecycle, NIL, IID_IShellFolder, Pointer(iRecycleFolder))) then
    Exit;
  End;
  ItemsAdded := 0;
  StartTime  := GetTickCount;
  Animate    := NIL;

  Try
  If SUCCEEDED(iRecycleFolder.EnumObjects(Self.Handle, SHCONTF_FOLDERS or SHCONTF_NONFOLDERS or SHCONTF_INCLUDEHIDDEN, EnumList)) Then
  Begin
     While EnumList.Next(1, PCurrList, Fetched) = S_OK do
     Begin
       FQPIDL := PIDL_Concatenate(PIDLRecycle, PCurrList);

       {Physical filename:}
       SetLength(FullPath, MAX_PATH);
       IF shGetPathFromIDList(FQPIDL, PChar(FullPath)) Then
         SetLength(FullPath, StrLen(PChar(FullPath)));

       {Filesize and -date:}
       If FindFirst(FullPath, faAnyFile, Srec) = 0 Then
       FindClose(Srec);
       Srec.Name := ExtractFilePath(FullPath) + Srec.Name;
       {SetLength(FullPath, 11);}

       FileSel := True;
       IF Assigned(OnAddFile) Then
        fOnAddFile(Self, Srec, FileSel);

       IF FileSel Then
       Begin
         {Displayname:}
         GetShellDisplayName(iRecycleFolder, PCurrList, SHGDN_FORPARSING, DisplayName);

         (*
         If SUCCEEDED(SHGetDataFromIDList(iRecycleFolder, PCurrList, SHGDFIL_FINDDATA, @Srec.FindData, SizeOf(SRec.FindData))) Then
           Srec.Size := Srec.FindData.nFileSizeLow;
         *)

         {Filetype and icon:}
         SHGetFileInfo(PChar(FQPIDL), 0, FileInfo, SizeOf(FileInfo), SHGFI_PIDL or SHGFI_TYPENAME or SHGFI_ICON);

         NewItem := AddItem(Srec);
         FileRec := NewItem.Data;
         FileRec^.Empty       := False;
         FileRec^.IconEmpty   := False;
         FileRec^.DisplayName := DisplayName;
         FileRec^.PIDL        := PCurrList;
         FileRec^.TypeName    := FileInfo.szTypeName;
         IF FileRec^.Typename = EmptyStr Then
         FileRec^.TypeName := Format(TextFileExt, [FileRec.FileExt]);
         FileRec^.ImageIndex  := FileInfo.iIcon;
         INC(ItemsAdded);
         IF ItemsAdded Mod 100 = 0 Then
         Begin
           Application.ProcessMessages;
           IF FShowAnimation And Not Assigned(Animate) And
              (GetTickCount - StartTime > AnimateDelay) And
               HandleAllocated And Not (csDesigning in ComponentState) Then
           Try
             Animate := TAnimate.Create(Self);
             Animate.Top  := (Height - Animate.Height) DIV 2;
             Animate.Left := (Width - Animate.Width) DIV 2;
             Animate.Parent := Self;
             Animate.CommonAVI := aviFindFolder;
             Animate.Active := True;
           Except Else
           End;
         End;
       End
       Else
         PIDL_Free(PCurrList);

       PIDL_Free(FQPIDL);
     End; {While EnumList ...}
  End;
  Finally

    If Assigned(PIDLRecycleLocal) Then
    PIDL_Free(PIDLRecycleLocal);

    If Assigned(Animate) Then
    Animate.Destroy;

    IF FShowMask Then
    UpdateNameHeader(False);
  End;

End; {LoadFromRecycleBin}



Function TDirView.GetNextMask(Var Mask : String) : String;
Var NextPos : Integer;
Begin
  NextPos := Pos(';', Mask);
  If NextPos = 0 Then
    Result := Mask
  Else
    Result := Copy(Mask, 1, Pred(NextPos));

  System.Delete(Mask, 1, Succ(Length(Result)));
End; {GetNextMask}


procedure TDirView.Load(Dir : String);
Var SRec            : TSearchRec;
    DosError        : Integer;
    TempMask        : String;
    ActMask         : String;
    ScanRun         : Integer;
    SaveCursor      : TCursor;
    FileSel         : Boolean;
    FileTimeFrom    : LongWord;
    FileTimeTo      : LongWord;
    FileList        : TStringList;
    Dummy           : Integer;
    AttrIncludeMask : Integer;
    AttrExcludeMask : Integer;
    SelTreeNode     : TTreeNode;
    ItemsAdded      : Integer;
    Animate         : TAnimate;
    StartTime       : DWORD;



begin {Load}
  IF Not FLoadEnabled Then
  Begin
    FDirty := True;
    Exit;
  End;

  fLoading      := True;
  fPath         := Dir;
  IF Assigned(fOnStartLoading) Then
  fOnStartLoading(Self);
  StopIconUpdateThread;
  StopWatchThread;

  fChangeTimer.Enabled  := False;
  FChangeTimer.Interval := 0;
  SaveCursor            := Screen.Cursor;
  Screen.Cursor         := crHourGlass;
  Animate               := NIL;

  GetSystemTime(fReloadTime);
  StartTime := GetTickCount;

  AttrIncludeMask := 0;
  AttrExcludeMask := 0;
  Case fSelArchive Of
    selYes: AttrIncludeMask := AttrIncludeMask Or faArchive;
    selNo:  AttrExcludeMask := AttrExcludeMask Or faArchive;
  End; {Case}

  Case fSelHidden Of
    selYes: AttrIncludeMask := AttrIncludeMask Or faHidden;
    selNo:  AttrExcludeMask := AttrExcludeMask Or faHidden;
  End; {Case}

  Case fSelReadOnly Of
    selYes: AttrIncludeMask := AttrIncludeMask Or faReadOnly;
    selNo:  AttrExcludeMask := AttrExcludeMask Or faReadOnly;
  End; {Case}

  Case fSelSysFile Of
    selYes: AttrIncludeMask := AttrIncludeMask Or faSysFile;
    selNo:  AttrExcludeMask := AttrExcludeMask Or faSysFile;
  End; {Case}

  FileTimeFrom := LongWord(fSelFileDateFrom) Shl 16 Or fSelFileTimeFrom;
  FileTimeTo   := LongWord(fSelFileDateTo)   Shl 16 Or fSelFileTimeTo;

  IF Assigned(DropTarget) Then
  DropTarget := NIL;

  Items.BeginUpdate;
  Items.Clear;
  Items.EndUpdate;
  Application.ProcessMessages;
  Items.BeginUpdate;
  FNotifyEnabled := False;
  FFilesSize    := 0;
  FFilesSelSize := 0;
  SortType      := stNone;

  ScanRun := 0;
  Try
    {ChDir(Dir);
    IF IOResult = 0 Then} {Don't change to displayed directory. Reason:
                           WINNT4 then refuses to move or delete this directory!}

    If (Length(Dir) > 0) Then
    Begin
      IF Not Assigned(FDriveView) Then
      DriveInfo.ReadDriveStatus(Dir[1], dsSize);
      fDriveType := DriveInfo[Dir[1]].DriveType;
    End
    Else
      fDriveType := DRIVE_UNKNOWN;

    IF FShowMask And FisRecycleBin Then
      UpdateNameHeader(FShowMask);

    fDirOK := (Length(Dir) > 0) And
             DriveInfo[Dir[1]].DriveReady And
             DirExists(Dir);

    IF fDirOK Then
    Begin
      IF Assigned(fDriveView) Then
      Begin
        SelTreeNode := TDriveView(fDriveView).FindNodeToPath(Dir);
        IF Assigned(SelTreeNode) Then
        fIsRecycleBin := TNodeData(SelTreeNode.Data).isRecycleBin
        Else
        fIsRecycleBin := Uppercase(Copy(Dir, 2, 10)) = ':\RECYCLED';
      End
      Else
        fIsRecycleBin := Uppercase(Copy(Dir, 2, 10)) = ':\RECYCLED';
{$IFNDEF PHYSICALRECYCLEBIN}
      IF FIsRecycleBin Then   
        LoadFromRecycleBin(Dir)
      Else
{$ENDIF}
      Begin
        TempMask := fMask;
        FileList := TStringList.Create;

        While Length(TempMask) > 0 Do
        Begin
          ActMask := GetNextMask(TempMask);
          INC(ScanRun);
          ItemsAdded := 0;

          IF Assigned(FileList) AND (Length(TempMask) > 0) Then
          FileList.Sort;

          DosError := SysUtils.FindFirst(AddSlash(Dir) + ActMask, FileAttr, SRec);
          While DosError = 0 Do
          Begin
            FileSel := True;
            IF (AttrIncludeMask <> 0) Then
            FileSel := Srec.Attr And AttrIncludeMask >= AttrIncludeMask;
            IF FileSel And (AttrExcludeMask <> 0) Then
            FileSel := AttrExcludeMask And Srec.Attr = 0;

            IF FileSel And
              ((SRec.Attr and faDirectory) = 0) And
              (Srec.Size >= fSelFileSizeFrom) And
              (Srec.Size <= fSelFileSizeTo)   And
              (LongWord(Srec.Time) >= FileTimeFrom) And
              (LongWord(Srec.Time) <= FileTimeTo)   Then
            Begin
              IF Assigned(OnAddFile) Then
              fOnAddFile(Self, Srec, FileSel);

              IF FileSel Then
              Begin
                IF (ScanRun = 1) OR
                   ((ScanRun > 1) And Not FileList.Find(SRec.Name, Dummy)) Then
                Begin
                  AddItem(Srec);
                  INC(ItemsAdded);
                  IF ItemsAdded Mod 100 = 0 Then
                  Begin
                    Application.ProcessMessages;
                    IF FShowAnimation And Not Assigned(Animate) And
                      (GetTickCount - StartTime > AnimateDelay) And
                       HandleAllocated And Not (csDesigning in ComponentState) Then
                    Try
                      Animate := TAnimate.Create(Self);
                      Animate.Top  := (Height - Animate.Height) DIV 2;
                      Animate.Left := (Width - Animate.Width) DIV 2;
                      Animate.Parent := Self;
                      Animate.CommonAVI := aviFindFolder;
                      Animate.Active := True;
                    Except Else
                    End;
                  End;
                  IF Length(TempMask) > 0 Then
                  FileList.Add(Srec.Name);
                End;
              End;
            End;
            DosError := FindNext(SRec);
          End;
          SysUtils.FindClose(SRec);
        End; {Length (TempMask) > 0}

        {Search for directories:}
        IF FShowDirectories Then
        Begin
          DosError := SysUtils.FindFirst(AddSlash(Dir) + '*.*', faDirectory or faSysFile or faHidden, SRec);
          While DosError = 0 Do
          Begin
            IF (Srec.Name <> '.') And (Srec.Name <> '..') And
              ((Srec.Attr And faDirectory) <> 0)          Then
            Begin
              FileSel := True;
              IF Assigned(OnAddFile) Then
              fOnAddFile(Self, Srec, FileSel);

              IF FileSel Then
                AddItem(Srec);
            End;
            DosError := FindNext(SRec);
          End;
          SysUtils.FindClose(SRec);
        End; {If FShowDirectories}

        IF Assigned(FileList) Then
        FileList.Free;
      End;
    End
    Else
      fisRecycleBin := False;
  Finally
    IF Assigned(Animate) Then
    Animate.Free;

    SetLength(ActMask, 0);
    fInfoCacheList.Sort(CompareInfoCacheItems);
    IF DirOK Then
    AlphaSort;
    fLoading := False;
    FNotifyEnabled := True;
    Items.EndUpdate;
    FDirty        := False;
    Screen.Cursor := SaveCursor;
  End;

  IF DirOK Then
  Begin
    IF fUseIconUpdateThread And Not FisRecycleBin Then
    StartIconUpdateThread;
    IF fWatchForChanges Then
    StartWatchThread;
  End;

  IF Assigned(fOnLoaded) Then
  fOnLoaded(Self);
end; {Load}


Procedure TDirView.ReLoad(CacheIcons : Boolean);
Var OldSel      : TStringList;
    i           : Integer;
    FoundIndex  : Integer;
    SaveItemFocused : String;
    ICacheItem  : TIconRec;
    PICacheItem : PIconRec;
    IconCache   : TListExt;

Begin
  IF (Length(fPath) = 0) Then
  Exit;

  IF Not FLoadEnabled Then
  Begin
    FDirty := True;
    Exit;
  End;

  Items.BeginUpdate;

  OldSel := TStringList.Create;
  IF CacheIcons Then
  IconCache := TListExt.Create(SizeOf(TIconRec))
  Else
  IConCache := NIL;

  For i := 0 To Pred(Items.Count) Do
  With Items[i] Do
  Begin
    IF Selected Then
    OldSel.Add(PFileRec(Data)^.FileName);
    IF CacheIcons And
       Not PFileRec(Data)^.IconEmpty And
       (Pos(PFileRec(Data)^.FileExt, SpecialExtensions) <> 0) Then
    Begin
      GetMem(PICacheItem, SizeOf(TIconRec));
      PICacheItem^.FileName   := PFileRec(Data)^.FileName;
      PICacheItem^.ImageIndex := PFileRec(Data)^.ImageIndex;
      IconCache.Add(PICacheItem);
    End;
  End;

  SaveItemFocused := '';
  IF Assigned(ItemFocused) Then
    SaveItemFocused := PFileRec(ItemFocused.Data)^.FileName;

  {Repopulate the listview:}
  Load(fPath);

  OldSel.Sort;

  IF CacheIcons Then
  IconCache.Sort(CompareIconCacheItems);

  {Restore the items state:}
  For i := 0 To Pred(Items.Count) Do
  With Items[i] Do
  Begin
    IF PFileRec(Data)^.FileName = SaveItemFocused Then
      ItemFocused := Items[i];

    IF OldSel.Find(PFileRec(Data)^.FileName, FoundIndex) Then
    Selected := True;

    IF CacheIcons And
       PFileRec(Data)^.IconEmpty Then
    Begin
      ICacheItem.FileName := PFileRec(Data)^.FileName;
      FoundIndex := IconCache.Find(Addr(ICacheItem), CompareIconCacheItems);
      IF FoundIndex >= 0 Then
      Begin
        PFileRec(Data)^.ImageIndex := PIconRec(IconCache.Data[FoundIndex])^.ImageIndex;
        PFileRec(Data)^.IconEmpty  := False;
      End;
    End;
  End;

  IF Assigned(ItemFocused) Then
  ItemFocused.MakeVisible(False);
  Items.EndUpdate;
  OldSel.Free;
  IF CacheIcons Then
  IconCache.Free;
End; {ReLoad}


Procedure TDirView.ClearIconCache;
Begin
  IF Assigned(FInfoCacheList) Then
  FInfoCacheList.Clear;
End; {ClearIconCache}


Function  TDirView.FormatFileTime(FileTime : TFILETIME) : String;
Var SysTime       : TSystemTime;
    DateTime      : TDateTime;
    LocalFileTime : TFILETIME;

Begin
  FileTimeToLocalFileTime(FileTime, LocalFileTime);
  FileTimeToSystemTime(LocalFileTime, SysTime);
  DateTime := SystemTimeToDateTime(SysTime);
  Result := FormatDateTime(DateTimeFormatStr, DateTime);
End; {FormatFileTime}


Function TDirView.GetAttrString(Attr : Integer) : String;
Begin
 Result := '';
 If Bool(Attr And FILE_ATTRIBUTE_READONLY)   Then Result := Result + 'R';
 If Bool(Attr And FILE_ATTRIBUTE_HIDDEN)     Then Result := Result + 'H';
 If Bool(Attr And FILE_ATTRIBUTE_SYSTEM)     Then Result := Result + 'S';
 If Bool(Attr And FILE_ATTRIBUTE_ARCHIVE)    Then Result := Result + 'A';
 If Bool(Attr And FILE_ATTRIBUTE_DIRECTORY)  Then Result := Result + 'D';
{If Bool(Attr And FILE_ATTRIBUTE_TEMPORARY)  Then Result := Result + 'T';}
 If Bool(Attr And FILE_ATTRIBUTE_COMPRESSED) Then Result := Result + 'C'; {NTFS only}


End; {GetAttrString}



Procedure TDirView.GetDisplayData(Item : TListItem; FetchIcon : Boolean);
Var FileInfo     : TShFileInfo;
    i            : Integer;
    PExtItem     : PInfoCache;
    CacheItem    : TInfoCache;
    IsSpecialExt : Boolean;

Begin
  with PFileRec(Item.Data)^ do
  Begin
    IsSpecialExt := MatchesFileExt(FileExt, SpecialExtensions);
    IF FUseIconCache And Not IsSpecialExt And Not isDirectory Then
    Begin
      CacheItem.FileExt := FileExt;
      i := FInfoCacheList.FindSequential(Addr(CacheItem), CompareInfoCacheItems);
      IF i >= 0 Then
      Begin
        TypeName    := PInfoCache(fInfoCacheList[i])^.TypeName;
        ImageIndex  := PInfoCache(fInfoCacheList[i])^.ImageIndex;
        Empty := False;
        IconEmpty := False;
      End;
    End;

    FetchIcon := IconEmpty And (FetchIcon Or Not IsSpecialExt);

    IF Empty Or FetchIcon Then
    Begin
      IF FetchIcon Then
      Begin
        If isDirectory Then
        Begin
          IF fDriveType = DRIVE_FIXED Then
          Begin
            Try
              {Retrieve icon and typename for the directory}
              shGetFileInfo(PChar(fPath + '\' + FileName), 0, FileInfo, SizeOf(FileInfo),
                            SHGFI_TYPENAME Or SHGFI_SYSICONINDEX Or SHGFI_ICON);
              IF (FileInfo.iIcon <= 0) Or (FileInfo.iIcon > SmallImages.Count) Then
              {Invalid icon returned: retry with access file attribute flag:}
              shGetFileInfo(PChar(fPath + '\' + FileName), FILE_ATTRIBUTE_DIRECTORY, FileInfo, SizeOf(FileInfo),
                            SHGFI_TYPENAME Or SHGFI_SYSICONINDEX Or SHGFI_USEFILEATTRIBUTES Or SHGFI_ICON);
              TypeName    := FileInfo.szTypeName;
              IF FetchIcon Then
              Begin
                ImageIndex  := FileInfo.iIcon;
                IconEmpty   := False;
              End;
            {Capture exceptions generated by the shell}
            Except Else
              Begin
                ImageIndex := StdDirIcon;
                IconEmpty  := False;
              End;
            End; {Except}
          End
          Else
          Begin
            TypeName   := StdDirTypeName;
            ImageIndex := StdDirIcon;
            IconEmpty  := False;
          End;
        End Else
        Begin
          {Retrieve icon and typename for the file}
          Try
            shGetFileInfo(PChar(fPath + '\' + FileName), FILE_ATTRIBUTE_NORMAL, FileInfo, SizeOf(FileInfo),
                          SHGFI_TYPENAME Or SHGFI_USEFILEATTRIBUTES Or SHGFI_SYSICONINDEX Or SHGFI_ICON);
            TypeName    := FileInfo.szTypeName;
            IF FetchIcon Then
            Begin
              ImageIndex  := FileInfo.iIcon;
              IconEmpty   := False;
            End;
          {Capture exceptions generated by the shell}
          Except Else
            Begin
              ImageIndex := UnKnownFileIcon;
              IconEmpty  := False;
            End;
          End; {Except}
        End;


        IF (Length(TypeName) > 0) Then
        Begin
          {Fill FileInfoCache:}
          IF FUseIconCache And Not IsSpecialExt And Not IconEmpty And Not isDirectory Then
          Begin
            GetMem(PExtItem, SizeOf(TInfoCache));
            PExtItem.FileExt    := FileExt;
            PExtItem.TypeName   := TypeName;
            PExtItem.ImageIndex := ImageIndex;
            fInfoCacheList.Add(PExtItem);
          End;
        End
        Else   {Length (TypeName = 0}
          TypeName := Format(TextFileExt, [FileExt]);
      End
      Else
      Begin
        Try
          shGetFileInfo(PChar(fPath + '\' + FileName), FILE_ATTRIBUTE_NORMAL, FileInfo, SizeOf(FileInfo),
                              SHGFI_TYPENAME Or SHGFI_USEFILEATTRIBUTES);
          TypeName    := FileInfo.szTypeName;
        Except Else TypeName := ''; End; {Capture exceptions generated by the shell}

        IF IconEmpty Then
        Begin
          IF FileExt = ExeExtension Then
          ImageIndex := DefaultExeIcon
          Else
          ImageIndex  := UnKnownFileIcon;
        End;
      End;
      Empty := False;
    End;
  End;
End; {GetDisplayData}


Function TDirView.GetFullFileName(Item : TListItem) : String;
Begin
  IF Assigned(Item) And Assigned(Item.Data) Then
  Begin
    IF Not FisRecycleBin Then
    Result := fPath + '\' + PFileRec(Item.Data)^.FileName
    Else
    Result := PFileRec(Item.Data)^.FileName;
  End
  Else
  Result := '';
End;


Function TDirView.GetFileName(Item : TListItem)     : String;
Begin
  Result := PFileRec(Item.Data)^.FileName;
End; {GetFileName}


Function TDirView.GetFileNameOnly(Item : TListItem)     : String;
Begin
  Result := PFileRec(Item.Data)^.FileName;
  SetLength(Result, Length(Result) - Length(GetFileExt(Item)));
End; {GetFileNameOnly}


Function TDirView.GetFileExt(Item : TListItem)     : String;
Begin
  Result := ExtractFileExt(PFileRec(Item.Data)^.FileName);
End; {GetFileExt}



Function TDirView.NormalizeMask(Mask : String) : String;
Var TempStr   : String;

Begin
  While (Length(Mask) > 0) And (Mask[1] = Space) Do
  System.Delete(Mask, 1, 1);
  While (Length(Mask) > 0) And (Mask[Length(Mask)] = Space) Do
  System.Delete(Mask, Length(Mask), 1);

  IF Length(Mask) = 0 Then
  Mask := '*.*';

  TempStr := Mask;
  StrTranslate(TempStr, ' ;,;');

  While Pos(';;', TempStr) <> 0 Do
  System.Delete(TempStr, Pos(';;', TempStr), 1);

  Result := LowerCase(TempStr);
End; {NormalizeMask}


Procedure TDirView.UpdateNameHeader(ShowMask : Boolean);
Var AddedMask : String;
Begin
  {Set the column header:}
  If ShowMask And (FMask <> '*.*') Then
  AddedMask :=   ' (' + FMask + ')'
  Else
  AddedMask := EmptyStr;
  IF Columns.Count > 0 Then
  Case SysLocale.PriLangID Of
    LANG_GERMAN:
      Columns[0].Caption := GERMAN_ColIDName + AddedMask;
    LANG_FRENCH:
      Columns[0].Caption := FRENCH_ColIDName + AddedMask;
    Else
      Columns[0].Caption := ENGLISH_ColIDName + AddedMask;
  End; {Case}
  SetColumnImages;
End; {UpdateNameHeader}


Procedure TDirView.SetMask(M : String);
Begin
  M := NormalizeMask(M);
  IF fMask <> M Then
  Begin
    fMask := M;
    ReLoad(True);
  End;

  IF HandleAllocated Then
  UpdateNameHeader(FShowMask);

End; {SetMask}


Procedure TDirView.BuildFileList(List : TStringList);
Var i : Integer;
Begin
  For i := 0 To Pred(Items.Count) Do
  Begin
    IF Items[i].Selected Then
      List.Add(GetFullFileName(Items[i]));
  End;
End; {BuildFileList}


Function TDirView.DeleteSelectedFiles(AllowUndo : Boolean) : Boolean;
Var ItemIndex     : Integer;
    i             : Integer;
    DirDeleted    : Boolean;
    FileOperator  : TFileOperator;

Begin
    DirDeleted := False;
    IF Assigned(fDriveView) Then
    TDriveView(fDriveView).StopWatchThread;
    StopWatchThread;
    Items.BeginUpdate;
    FileOperator := TFileOperator.Create(Self);
    Try
      ItemIndex := Selected.Index;
      FileOperator.Operation     := foDelete;
      FileOperator.Flags         := [foNoConfirmMkDir];
      FileOperator.ProgressTitle := coFileOperatorTitle;
      BuildFileList(FileOperator.OperandFrom);

      IF FConfirmDelete Then
        FileOperator.Flags := FileOperator.Flags - [foNoConfirmation]
      Else
        FileOperator.Flags := FileOperator.Flags + [foNoConfirmation];

      IF AllowUndo Then
        FileOperator.Flags := FileOperator.Flags + [foAllowUndo]
      Else
        FileOperator.Flags := FileOperator.Flags - [foAllowUndo];

      StopIconUpdateThread;
      Result := FileOperator.Execute;
      Result := Result And Not FileOperator.OperationAborted;

      {Verify, wether the selected items still exists:}
      For i := Pred(Items.Count) DownTo 0 Do
      With Items[i] Do
      Begin
        IF Selected Then
        Begin
          Case PFileRec(Data)^.isDirectory Of
            True:  Begin
                     DirDeleted := True;
                     If Not DirExists(GetFullFileName(Items[i])) Then
                     Items[i].Delete;
                   End;
            False: IF Not FileExists(GetFullFileName(Items[i])) Then
                     Items[i].Delete;
          End; {Case}
          IF SelCount = 0 Then
          Break;
        End;
      End;
    Finally
      Items.EndUpdate;
      FileOperator.Free;
    End;

    IF fUseIconUpdateThread Then
    StartIconUpdateThread;

    IF fWatchForChanges Then
    StartWatchThread;

    IF Assigned(fDriveView) Then
    With TDriveView(fDriveView) Do
    Begin
      IF DirDeleted And Assigned(Selected) Then
         ValidateDirectory(Selected);
      TDriveView(fDriveView).StartWatchThread;
    End;

    IF (ItemIndex >= Items.Count) And (Items.Count > 0) Then
      Selected := Items[Pred(Items.Count)];
End; {DeleteSelectedFiles}


Function CompareFileName (I1, I2: TListItem; Ascending : Integer): Integer; StdCall;
Var F1, F2 : String;
Begin
  if I1 = I2 then  Result := 0  else
  if I1 = NIL then Result := -1 else
  if I2 = NIL then Result := 1  else
  begin
    {Directories allways should appear "grouped":}
    IF PFileRec(I1.Data).isDirectory <> PFileRec(I2.Data).isDirectory Then
    Begin
      IF PFileRec(I1.Data).isDirectory Then
      Result := fLess
      Else
      Result := fGreater;
    End
    Else
    Begin
      F1 := Uppercase(PFileRec(I1.Data)^.DisplayName);
      F2 := Uppercase(PFileRec(I2.Data)^.DisplayName);
      IF F1 < F2 Then Result := fLess    Else
      IF F1 > F2 Then Result := fGreater Else
                      Result := fEqual;
    End;
  End;

  IF Not Bool(Ascending) Then
  Result := -Result;
End; {CompareFileName}




Function CompareFileSize(I1, I2: TListItem; Ascending : Integer): Integer;  StdCall;
Begin
  if I1 = I2 then  Result := 0  else
  if I1 = NIL then Result := -1 else
  if I2 = NIL then Result := 1  else
  begin
    {Directories allways should appear "grouped":}
    IF PFileRec(I1.Data).isDirectory <> PFileRec(I2.Data).isDirectory Then
    Begin
      IF PFileRec(I1.Data).isDirectory Then
      Result := fLess
      Else
      Result := fGreater;
    End
    Else
    Begin
      IF PFileRec(I1.Data)^.Size < PFileRec(I2.Data)^.Size Then Result := fLess    Else
      IF PFileRec(I1.Data)^.Size > PFileRec(I2.Data)^.Size Then Result := fGreater Else
                                                Result := fEqual;
    End;
  End;
  IF Not Bool(Ascending) Then
  Result := -Result;
End; {CompareFileSize}


Function CompareFileType (I1, I2: TListItem; Ascending : Integer): Integer; StdCall;
Begin
  if I1 = I2 then  Result := 0  else
  if I1 = NIL then Result := -1 else
  if I2 = NIL then Result := 1  else
  begin
    IF PFileRec(I1.Data)^.Empty Then
    TDirView(I1.ListView).GetDisplayData(I1, False {Not TDirView(I1.ListView).FUseIconUpdateThread});
    IF PFileRec(I2.Data)^.Empty Then
    TDirView(I2.ListView).GetDisplayData(I2, False {Not TDirView(I2.ListView).FUseIconUpdateThread});

    {Directories allways should appear "grouped":}
    IF PFileRec(I1.Data).isDirectory <> PFileRec(I2.Data).isDirectory Then
    Begin
      IF PFileRec(I1.Data).isDirectory Then
      Result := fLess
      Else
      Result := fGreater;
    End
    Else
    Begin
      IF PFileRec(I1.Data)^.TypeName < PFileRec(I2.Data)^.TypeName Then Result := fLess    Else
      IF PFileRec(I1.Data)^.TypeName > PFileRec(I2.Data)^.TypeName Then Result := fGreater Else
                                                        Result := fEqual;
    End;
  End;
  IF Not Bool(Ascending) Then
  Result := -Result;
End; {CompareFileSize}


Function CompareFileAttr (I1, I2: TListItem; Ascending : Integer): Integer; StdCall;
Begin
  if I1 = I2 then  Result := 0  else
  if I1 = NIL then Result := -1 else
  if I2 = NIL then Result := 1  else
  begin
    {Directories allways should appear "grouped":}
    IF PFileRec(I1.Data).isDirectory <> PFileRec(I2.Data).isDirectory Then
    Begin
      IF PFileRec(I1.Data).isDirectory Then
      Result := fLess
      Else
      Result := fGreater;
    End
    Else
    Begin
      IF PFileRec(I1.Data)^.Attr < PFileRec(I2.Data)^.Attr Then Result := fLess    Else
      IF PFileRec(I1.Data)^.Attr > PFileRec(I2.Data)^.Attr Then Result := fGreater Else
                                              Result := fEqual;
    End;
  End;
  IF Not Bool(Ascending) Then
  Result := -Result;
End; {CompareFileAttr}


Function CompareFileTime (I1, I2: TListItem; Ascending : Integer): Integer;  StdCall;
Var Time1, Time2 : Int64;
Begin
  if I1 = I2 then  Result := 0  else
  if I1 = NIL then Result := -1 else
  if I2 = NIL then Result := 1  else
  begin
    Time1 := Int64(PFileRec(I1.Data)^.FileTime.dwHighDateTime) Shl 32 + PFileRec(I1.Data)^.FileTime.dwLowDateTime;
    Time2 := Int64(PFileRec(I2.Data)^.FileTime.dwHighDateTime) Shl 32 + PFileRec(I2.Data)^.FileTime.dwLowDateTime;
    {Directories allways should appear "grouped":}
    IF PFileRec(I1.Data).isDirectory <> PFileRec(I2.Data).isDirectory Then
    Begin
      IF PFileRec(I1.Data).isDirectory Then
      Result := fLess
      Else
      Result := fGreater;
    End
    Else
    Begin
      IF Time1 < Time2 Then Result := fLess    Else
      IF Time1 > Time2 Then Result := fGreater Else
                            Result := fEqual;
    End;
  End;
  IF Not Bool(Ascending) Then
  Result := -Result;
End; {CompareFileTime}



Function TDirView.CustomSort(SortProc: TLVCompare; Data: Longint): Boolean;
Begin
  Result := False;
  If HandleAllocated Then
  Begin
    If Not Assigned(SortProc) Then
    SortProc := @CompareFileName;
    Result := Inherited CustomSort(SortProc, Data);
  End;
End; {CustomSort}


Function TDirView.AlphaSort: Boolean;
Var CompareProc : TLVCompare;

Begin
  If HandleAllocated Then
  Begin
    Case SortColumn Of
    sName: CompareProc := @CompareFilename;
    sSize: CompareProc := @CompareFileSize;
    sType: CompareProc := @CompareFileType;
    sAttr: CompareProc := @CompareFileAttr;
    sDate: CompareProc := @CompareFileTime;
    Else
       CompareProc := @CompareFilename;
    End; {Case}
    StopIconUpdateThread;
    FNotifyEnabled := False;
    Result := CustomSort(CompareProc, Integer(fSortAscending));
    FNotifyEnabled := True;
    IF Not fLoading And fUseIconUpdateThread Then
    StartIconUpdateThread;
  End
  Else
    Result := False;
End; {AlphaSort}


Procedure TDirView.SetSortColumn(WichCol : TSortColumn);
Begin
  fSortColumn := Ord(WichCol);
  AlphaSort;
  SetColumnImages;
End; {SetSortColumn}


Function TDirView.GetSortColumn : TSortColumn;
Begin
  Result := TSortColumn(fSortColumn);
End; {GetSortColumn}


Procedure TDirView.SetSortAscending(Ascending : Boolean);
Begin
  fSortAscending := Ascending;
  AlphaSort;
  SetColumnImages;
End; {SetSortAscending}


Procedure TDirView.ValidateFile(Item : TListItem);
Var i : Integer;
Begin
  IF Not Assigned(Item) Or Not Assigned(Item.Data) Then
  Exit;
  i := Item.Index;
  IF Not FileExists(GetFullFileName(Items[i])) Then
    Item.Delete;
End; {ValidateFile}


Procedure TDirView.ValidateFile(FileName : TFileName);
Var FilePath : String;
Begin
  FilePath := RemoveSlash(ExtractFilePath(FileName));
  If FisRecycleBin Then
  ValidateFile(FindFileItem(FileName))
  Else IF (FilePath = FPath) Then
  ValidateFile(FindFileItem(ExtractFileName(FileName)));
End; {ValidateFile}


Procedure TDirView.ValidateSelectedFiles;
Var i : Integer;
    Upd : Boolean;

Begin
  i := 0;
  Upd := False;
  While i <= Pred(Items.Count) Do
  Begin
    IF Items[i].Selected And
    ((Not PFileRec(Items[i].Data)^.isDirectory And Not FileExists(GetFullFileName(Items[i])) Or
    (PFileRec(Items[i].Data)^.isDirectory And Not DirExists(GetFullFileName(Items[i]))))) Then
    Begin
      IF (SelCount > 10) And Not Upd Then
      Begin
        Items.BeginUpdate;
        Upd := True;
      End;
      Items[i].Delete;
    End
    Else
    INC(i);
  End;
  IF Upd Then
  Items.EndUpdate;
End; {ValidateSelectedFiles}


Function TDirView.CreateFile(NewName : String) : TListItem;
Var F         : File;
    Srec      : TSearchRec;

Begin
  Result := NIL;
  {Neue Datei anlegen:}
  NewName := Path + '\' + NewName;

  {Ermitteln des neuen Dateinamens:}
  IF FileExists(NewName) Then
  Exit;

  IF FWatchForChanges Then
  StopWatchThread;

  StopIconUpdateThread;
  Try

  {Create the desired file as empty file:}
    AssignFile(F, NewName);
    Rewrite(F);
    IF IOResult = 0 Then
    Begin
      CloseFile(F);

      {Anlegen der Datei als TListItem:}
      IF FindFirst(NewName, faAnyFile, Srec) = 0 Then
      Begin
        Result := AddItem(SRec);
        ItemFocused := FindFileItem(GetFileRec(Result.Index)^.FileName);
        ItemFocused.MakeVisible(False);
      End;
      FindClose(Srec);
    End;
  Finally
    If FUseIconUpdateThread Then
    StartIconUpdateThread;

    IF fWatchForChanges Then
    StartWatchThread;
  End;
End; {CreateFile}


Procedure TDirView.SetColumnImages;
var Hditem : THdItem;
    i      : Integer;
Begin
  IF Not ShowColumnHeaders Then
  Exit;

  for i := 0 to Columns.Count-1 do
  begin
    IF FBitmap.Handle <> NULL Then
    FBitmap.Releasehandle;

    Hditem.Mask := HDI_FORMAT;
    Header_GetItem(GetDlgItem(Self.Handle,0), I, Hditem);
    Hditem.Mask := HDI_BITMAP or HDI_FORMAT;

    if fShowColumnIcon And (i = fSortColumn) Then
    Begin
      if fSortAscending then
        FBitmap.Handle := ArrowDown
      else
        FBitmap.Handle := ArrowUp;
      Hditem.fmt := Hditem.fmt Or HDF_BITMAP;

      IF Columns[i].Alignment = taLeftJustify Then
       Hditem.fmt := Hditem.fmt Or HDF_BITMAP_ON_RIGHT;
    End
    Else
    Begin
      FBitmap.Handle := ArrowDown;
      Hditem.fmt := Hditem.fmt And Not (HDF_BITMAP or HDF_BITMAP_ON_RIGHT);
    End;
    Hditem.hbm := FBitmap.Handle;
    Header_SetItem(GetDlgItem(Self.Handle,0),I,hditem);
  End;
  FColumnIconPainted := True;
End; {SetColumnImages}


Procedure TDirView.SetColumnProperties;
Begin
  IF Columns.Count < 4 Then
  Exit;
  Columns.BeginUpdate;

  SetMask(FMask);

  Case SysLocale.PriLangID Of
    LANG_GERMAN:
    Begin
      Columns[1].Caption := GERMAN_ColIDSize;
      Columns[2].Caption := GERMAN_ColIDType;
      Columns[3].Caption := GERMAN_ColIDDate;
      Columns[4].Caption := GERMAN_ColIDAttr;
    End;
    LANG_FRENCH:
    Begin
      Columns[1].Caption := FRENCH_ColIDSize;
      Columns[2].Caption := FRENCH_ColIDType;
      Columns[3].Caption := FRENCH_ColIDDate;
      Columns[4].Caption := FRENCH_ColIDAttr;
    End;
    Else
    Begin
      Columns[1].Caption := ENGLISH_ColIDSize;
      Columns[2].Caption := ENGLISH_ColIDType;
      Columns[3].Caption := ENGLISH_ColIDDate;
      Columns[4].Caption := ENGLISH_ColIDAttr;
    End;
  End; {Case}

  {Set column width:}
  Columns[0].Width := FColWidthName;
  Columns[1].Width := FColWidthSize;
  Columns[2].Width := FColWidthType;
  Columns[3].Width := FColWidthDate;
  Columns[4].Width := FColWidthAttr;

  Columns[0].Alignment := taLeftJustify;
  Columns[1].Alignment := taRightJustify;
  Columns[2].Alignment := taLeftJustify;
  Columns[3].Alignment := taLeftJustify;
  Columns[4].Alignment := taRightJustify;
  Columns.EndUpdate;
End; {SetColumnProperties}


Procedure TDirView.CreateColumns;
Var i : Integer;
Begin
  Columns.BeginUpdate;
  If Columns.Count > 0 Then
  Columns.Clear;
  For i := 0 To 4 Do
  Columns.Add;
  SetColumnProperties;
  Columns.EndUpdate;
End; {CreateColumns}


Procedure TDirView.SaveColumnProperties;
Begin
  IF Columns.Count > 0 Then
    ColWidthName := Columns[0].Width;
  IF Columns.Count > 1 Then
    ColWidthSize := Columns[1].Width;
  IF Columns.Count > 2 Then
    ColWidthType := Columns[2].Width;
  IF Columns.Count > 3 Then
    ColWidthDate := Columns[3].Width;
  IF Columns.Count > 4 Then
    ColWidthAttr := Columns[4].Width;
End; {SaveColumnProperties}



Procedure TDirView.DisplayPropertiesMenu;
Var FileList : TStringList;
    i        : Integer;

Begin
  IF (SelCount = 1) Or FisRecycleBin Then
  Begin
    If FisRecycleBin Then
    ShellExecuteContextCommand(FParentForm.Handle, shcProperties, iRecycleFolder, 1, PFileRec(Selected.Data)^.PIDL )
    Else
    ShellExecuteContextCommand(FParentForm.Handle, shcProperties, GetFullFileName(Selected));
  End
  Else
  IF SelCount > 1 Then
  Begin
    FileList := TStringList.Create;
    BuildFileList(FileList);
    For i := 0 to Pred(FileList.Count) Do
    FileList[i] := ExtractFileName(FileList[i]);
    ShellExecuteContextCommand(FParentForm.Handle, shcProperties, PathName, FileList);
    FileList.Destroy;
  End;
End; {DisplayProperiesMenu}


Procedure TDirView.DisplayContextMenu;
Var FileList  : TStringList;
    i         : Integer;
    DefDir    : String;
    Verb      : String;
    PIDLArray : PPIDLArray;
    Count     : Integer;
    DiffSelectedPath : Boolean;

Begin
    GetDir(0, DefDir);
    ChDir(PathName);
    Verb := EmptyStr;
    IF SelCount = 1 Then
    Begin
      Verb := EmptyStr;
      IF FisRecycleBin Then
        ShellDisplayContextMenu(FParentForm.Handle, Mouse.CursorPos, iRecycleFolder, 1, PFileRec(Selected.Data)^.PIDL , False, Verb)
      Else
      Begin
        ShellDisplayContextMenu(FParentForm.Handle, Mouse.CursorPos, GetFullFileName(Selected), CanEdit(Selected), Verb);
        LoadEnabled := True;
      End; {not FisRecycleBin}

      If Verb = 'rename' Then
      Selected.EditCaption
      Else If Verb = shcCut Then
      Begin
        For i := 0 To Items.Count - 1 Do
        IF Items[i].Cut Then
        Items[i].Cut := False;
        Selected.Cut := True;
      End;
    End
    Else
    IF SelCount > 1 Then
    Begin
      If FisRecycleBin Then
      Begin
        Count := 0;
        GetMem(PIDLArray, SizeOf(PItemIDList) * SelCount);
        FillChar(PIDLArray^, Sizeof(PItemIDList) * SelCount, #0);
        For i := Selected.Index To Items.Count - 1 Do
        IF Items[i].Selected Then
        Begin
          PIDLArray^[Count] := PFileRec(Items[i].Data)^.PIDL;
          INC(Count);
        End;
        ShellDisplayContextMenu(FParentForm.Handle, Mouse.CursorPos, iRecycleFolder, Count, PidlArray^[0], False, Verb);
        FreeMem(PIDLArray, Count);
      End
      Else
      Begin
        FileList := TStringList.Create;
        BuildFileList(FileList);
        For i := 0 To FileList.Count - 1 Do
        FileList[i] := ExtractFileName(FileList[i]);
        ShellDisplayContextMenu(FParentForm.Handle, Mouse.CursorPos, PathName, FileList, Verb);
        FileList.Destroy;
      End;

      If Verb = shcCut Then
      Begin
        LastClipBoardOperation := cboCut;
        For i := 0 To Items.Count - 1 Do
        With Items[i] Do
        IF Selected Then
        Cut := True
        Else IF Cut Then
        Items[i].Cut := False;
      End
      Else If Verb = shcCopy Then
      LastClipBoardOperation := cboCopy;
    End;
    ChDir(DefDir);


    If (FisRecycleBin And (Verb <> shcCut) And (Verb <> shcProperties)) Then
    Begin
      DiffSelectedPath := False;
      For i := Selected.Index To Items.Count - 1 Do
      IF ExtractFilePath(PFileRec(Items[i].Data)^.FileName) <> FPath + '\' Then
      Begin
        DiffSelectedPath := True;
        Break;
      End;
      IF DiffSelectedPath Then
      Begin
        DragFileList.Clear;
        BuildFileList(DragFileList);
        TFileDeleteThread.Create(DragFileList, MaxWaitTimeOut, SignalFileDelete);
        Exit;
      End;
    End;

    IF Not WatchThreadActive Then
    Begin
      Sleep(500);
      ValidateSelectedFiles;
    End;
End; {ContextMenu}


Procedure TDirView.SelectFiles(Filter : TFileFilter; Select : Boolean);
Var SaveCursor      : TCursor;
    Srec            : TSearchRec;
    DosError        : Integer;
    Mask            : String;
    FileSel         : Boolean;
    FileTimeFrom    : LongWord;
    FileTimeTo      : LongWord;
    AttrIncludeMask : Integer;
    AttrExcludeMask : Integer;
    Files           : TStringList;
    i               : Integer;

Begin
  Files := TStringList.Create;
  For i := 0 To Items.Count - 1 Do
  Files.AddObject(PFileRec(Items[i].Data)^.FileName, Pointer(i));
  Files.Sort;

  SaveCursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;

  FileTimeFrom := LongWord(Filter.SelFileDateFrom) Shl 16 Or Filter.SelFileTimeFrom;
  FileTimeTo   := LongWord(Filter.SelFileDateTo)   Shl 16 Or Filter.SelFileTimeTo;

  AttrIncludeMask := 0;
  AttrExcludeMask := 0;
  Case Filter.SelArchive Of
    selYes: AttrIncludeMask := AttrIncludeMask Or faArchive;
    selNo:  AttrExcludeMask := AttrExcludeMask Or faArchive;
  End; {Case}

  Case Filter.SelHidden Of
    selYes: AttrIncludeMask := AttrIncludeMask Or faHidden;
    selNo:  AttrExcludeMask := AttrExcludeMask Or faHidden;
  End; {Case}

  Case Filter.SelReadOnly Of
    selYes: AttrIncludeMask := AttrIncludeMask Or faReadOnly;
    selNo:  AttrExcludeMask := AttrExcludeMask Or faReadOnly;
  End; {Case}

  Case Filter.SelSysFile Of
    selYes: AttrIncludeMask := AttrIncludeMask Or faSysFile;
    selNo:  AttrExcludeMask := AttrExcludeMask Or faSysFile;
  End; {Case}

  Items.BeginUpdate;

  While Length(Filter.Mask) > 0 Do
  Begin
    Mask := GetNextMask(Filter.Mask);
    DosError := SysUtils.FindFirst(FPath + '\' + Mask, FileAttr, Srec);
    While DosError = 0 Do
    Begin
      FileSel := True;
      IF FileSel And
        (Abs(Srec.Size) >= Filter.SelFileSizeFrom) And
        (Abs(Srec.Size) <= Filter.SelFileSizeTo)   And
        (LongWord(Srec.Time) >= FileTimeFrom) And
        (LongWord(Srec.Time) <= FileTimeTo)   Then
      Begin
        IF (AttrIncludeMask <> 0) Then
        FileSel := Srec.Attr And AttrIncludeMask >= AttrIncludeMask;
        IF FileSel And (AttrExcludeMask <> 0) Then
        FileSel := AttrExcludeMask And Srec.Attr = 0;

        IF FileSel Then
        Begin
          If (Files.Find(Srec.Name, i) And (i < Items.Count)) Then
            Items[Integer(Files.Objects[i])].Selected := Select;
        End;
     End;
     DosError := SysUtils.FindNext(Srec);
    End;
    SysUtils.FindClose(SRec);
  End;

  Items.EndUpdate;
  Files.Destroy;
  Screen.Cursor := SaveCursor;
End; {SelectFiles}


procedure TDirView.CNNotify(var Msg: TWMNotify);
Var Item:TListItem;
    InfoMask : UINT;

Begin

  IF (Msg.NMHdr.code = LVN_GETDISPINFO) And FNotifyEnabled Then
  Begin
    Try
      With PLVDispInfo(Pointer(Msg.NMHdr))^.item do
      Begin
        InfoMask := PLVDispInfo(Pointer(Msg.NMHdr))^.item.Mask;

        If (Infomask And LVIF_PARAM) <> 0 Then
          Item := TListItem(lParam)
        else
          Item := Items[IItem];

        IF Assigned(Item) And Assigned(Item.Data) Then
        Begin
          With PFileRec(Item.Data)^ do
          Begin
            {Fetch display data of current file:}
            IF Empty Then
              GetDisplayData(Item, IconEmpty And (Not fUseIconUpdateThread
                              Or ((ViewStyle <> vsReport) And (Win32PlatForm = VER_PLATFORM_WIN32_NT))));

            IF IconEmpty And
               (Not fUseIconUpdateThread
                Or ((ViewStyle <> vsReport) And (Win32PlatForm = VER_PLATFORM_WIN32_NT))) And
               ((Infomask and LVIF_IMAGE) <> 0) Then
              GetDisplayData(Item, True);

              {Set IconUpdatethread :}
              IF IconEmpty And Assigned(FIconUpdateThread) Then
              Begin
                IF Assigned(TopItem) Then
                {Viewstyle is vsReport or vsList:}
                fIconUpdateThread.Index := Self.TopItem.Index
                Else
                {Viewstyle is vsIcon or vsSmallIcon:}
                fIconUpdateThread.MaxIndex := Item.Index;

                IF fIconUpdateThread.Suspended And Not FisRecycleBin Then
                fIconUpdateThread.Resume;
              End;

            If (Infomask and LVIF_TEXT) <> 0 then
            Begin
              If iSubItem = 0 then
                StrPLCopy(pszText, DisplayName, cchTextMax)
              Else
              With Item.SubItems do
              Begin
                If iSubItem <= Count then
                  Case iSubItem of
                    1: {Size:     }
                       IF Not isDirectory Then
                         StrPLCopy(pszText, FormatSize(Size), cchTextMax);
                    2: {FileType: }
                        StrPLCopy(pszText, TypeName, cchTextMax);
                    3: {Date}
                        StrPLCopy(pszText,FormatFileTime(FileTime), cchTextMax);
                    4: {Attrs:    }
                        StrPLCopy(pszText, GetAttrString(Attr), cchTextMax);
                  End { case }
                Else
                  pszText[0] := #0;
              End; {SubItem}
            End; {With PFileRec(Item.Data)^ Do}

            {Set display icon of current file:}
            If (Infomask And LVIF_IMAGE) <> 0 Then
              iImage := PFileRec(Item.Data).ImageIndex;
          End; {With PFileRec Do}
          {mask := mask or LVIF_DI_SETITEM; <== causes icons not to be updated on renaming the item}
        End; {Assigned(Item}
      End; {With PLVDispInfo}
    Except Else End; {Capture exceptions}
  End {Message LVN_GETDISPINFO}
  Else
  Begin
    Inherited;
    Case Msg.NMHdr.code Of
      LVN_ENDLABELEDIT: LoadEnabled := True;
{$IFDEF DRAGDROP}
      LVN_BEGINDRAG:    DDDragDetect(MK_LBUTTON, FStartPos, Mouse.CursorPos, ddsDrag);
      LVN_BEGINRDRAG:   DDDragDetect(MK_RBUTTON, FStartPos, Mouse.CursorPos, ddsDrag);
{$ENDIF}
    End; {Case}
  End;
End; {CNNotify}


{Repaint header icons, if columns are resized:}
Procedure TDirView.WMNotify(var msg: TWMNotify);
Begin
  IF (fHeaderHandle <> 0)                  And
     (Msg.NMHdr^.hWndFrom = FHeaderHandle) Then
    Case Msg.NMHdr.code Of
      HDN_ENDDRAG: Begin
                     {Due to a bug in D4 (until Update Pack 3) we should eat this message:}
                     Invalidate;
                     Exit;
                   End;
    HDN_ENDTRACK, HDN_ENDTRACKW:
                   Begin
                     Inherited;
                     SetColumnImages;
                     FColumnIconPainted := False;
                     Exit;
                   End;
  End; {Case}
  Inherited;
End; { TDirView.WMNotify }


Procedure TDirView.WMPaint(Var Msg : TWMPaint);
Begin
  Inherited;
  IF (ViewStyle = vsReport) And Not
     FColumnIconPainted     And
     (ShowColumnHeaders)    Then
     SetColumnImages;
End; {TDirView.WMPaint}


Procedure TDirView.CreateWnd;
Begin
  Inherited CreateWnd;
  IconsSetImageList;

  IF Assigned(PopupMenu) Then
  PopupMenu.Autopopup := False;
  FParentForm := GetParentForm(Self);
  IF Not Assigned(DragImageList) And Assigned(FParentForm) And Not (csDesigning in ComponentState) Then
  DragImageList := TImageList.Create(FParentForm);
  SetMask(FMask);
End; {CreateWnd}


Procedure TDirView.InternalOnCustomDrawItem(Sender: TCustomListView;
                                            Item: TListItem;
                                            State: TCustomDrawState;
                                            var DefaultDraw: Boolean);
Begin
  IF fDimmHiddenFiles And Not Item.Selected And
      (PFileRec(Item.Data).Attr And faHidden <> 0) Then
       Canvas.Font.Color := clGrayText;
End; {CustomDrawItem}


Procedure TDirView.InternalOnCustomDrawSubItem(Sender: TCustomListView;
                                               Item: TListItem;
                                               SubItem: Integer;
                                               State: TCustomDrawState;
                                               var DefaultDraw: Boolean);
begin
  if fDimmHiddenFiles And
     (SubItem > 0)    then
    Canvas.Font.Color := GetSysColor(COLOR_WINDOWTEXT);
end; {InternalOnCustomDrawSubItem}


Procedure TDirView.StartIconUpdateThread;
Begin
  IF fDirOK Then
  Begin
    IF Not Assigned(fIconUpdateThread) Then
    Begin
      If Items.Count > 0 Then
      fIconUpdateThread  := TIconUpdateThread.Create(Self);
    End
    Else
    Begin
      Assert(Not fIConUpdateThread.Terminated);
      fIconUpdateThread.Index := 0;
      IF ViewStyle = vsReport Then
      fIconUpdateThread.Resume;
    End;
  End;
End; {StartIconUpdateThread}


Procedure TDirView.StopIconUpdateThread;
Var Counter : Integer;
Begin
  IF Assigned(fIconUpdateThread) Then
  Begin
    Counter := 0;
    fIconUpdateThread.Terminate;
    fIconUpdateThread.Priority := tpHigher;
    IF fIconUpdateThread.Suspended Then
    fIconUpdateThread.Resume;
    Sleep(0);

    Try
      {Wait until the thread has teminated to prevent AVs:}
      While Not fIUThreadFinished Do
      Begin
        Sleep(10);
        Application.ProcessMessages;
        INC(Counter);
        {Raise an exception after 2 second, if the thread has not terminated:}
        IF Counter = 200 Then
        Begin
          Raise EIUThread.Create('Can''t terminate icon update thread.');
          Break;
        End;
      End;
    Finally
      fIconUpdateThread.Destroy;
      fIconUpdateThread := NIL;
    End;
  End;
End; {StopIconUpdateThread}


procedure TDirView.StopWatchThread;
Begin
  if Assigned(FDiscMonitor) then
    FDiscMonitor.Active := False;
End; {StopWatchThread}


procedure TDirView.StartWatchThread;
Begin
  IF (Length(FPath) = 0) Or Not FWatchForChanges Or Not fDirOK Then
    Exit;

  IF Not Assigned(FDiscMonitor) Then
  Begin
    FDiscMonitor := TDiscMonitor.Create(Self);
    With FDiscMonitor Do
    Begin
      ChangeDelay := msThreadChangeDelay;
      SubTree   := False;
      Filters   := [moDirName, moFileName, moSize, moAttributes];
      Directory := PathName;
      OnChange  := ChangeDetected;
      OnInvalid := ChangeInvalid;
      Open;
    End;
  End
  Else
  Begin
    FDiscMonitor.Directory := PathName;
    FDiscMonitor.Open;
  End;
End; {StartWatchThread}


Procedure TDirView.TimerOnTimer(Sender : TObject);
Begin
  IF Not fLoading Then
  Begin
    IF Assigned(fOnChangeDetected) Then
    fOnChangeDetected(Self);
    Reload(True);
  End
End; {TimerOnTimer}


procedure TDirView.ChangeDetected(Sender: TObject);
Begin
  FDirty := True;
  FChangeTimer.Enabled := False;
  FChangeTimer.Interval := 0;
  FChangeTimer.Interval := fChangeInterval;
  FChangeTimer.Enabled := True;
End; {ChangeDetected}


procedure TDirView.ChangeInvalid(Sender: TObject);
Begin
  fDiscMonitor.Close;
  IF Assigned(fOnChangeInvalid) Then
  fOnChangeInvalid(Self);
End; {ChangeInvalid}


Procedure TDirView.Syncronize;
Begin
  Application.ProcessMessages;
  FChangeTimer.Enabled  := False;
  FChangeTimer.Interval := 0;
  LoadEnabled := True;
  IF FDirty Then
  Reload(True);
End; {Syncronize}


Function TDirView.WatchThreadActive : Boolean;
Begin
  Result := FWatchForChanges And Assigned(FDiscMonitor) And FDiscMonitor.Active;
End; {WatchThreadActive}


Procedure TDirView.SetChangeInterval(Interval : Cardinal);
Begin
  IF Interval > 0 Then
  Begin
    fChangeInterval := Interval;
    fChangeTimer.Interval := Interval;
  End;
End; {SetChangeInterval}


Procedure TDirView.SetShowColumnIcon(ShowIt : Boolean);
Begin
  IF ShowIt = fShowColumnIcon Then
  Exit;
  fShowColumnIcon := ShowIt;
  SetColumnImages;
End; {SetShowColumnIcon}

Procedure TDirView.SetViewStyle(VStyle : TViewStyle);
Begin
  IF VStyle <> ViewStyle Then
  Begin
    Inherited ViewStyle := VStyle;
    SetColumnImages;
  End;
End;

Procedure TDirView.SetDimmHiddenFiles(Dimm : Boolean);
Begin
  IF Dimm = fDimmHiddenFiles Then
  Exit;
  fDimmHiddenFiles := Dimm;
  Self.Repaint;
End; {SetDimmHiddenFiles}


Procedure TDirView.SetShowDirectories(ShowIt : Boolean);
Begin
  IF ShowIt <> fShowDirectories Then
  Begin
    fShowDirectories := ShowIt;
    Reload(True);
  End;
End; {SetShowDirectories}


Procedure TDirView.SetDateTimeDisplay(Format: TDateTimeDisplay);
Begin
  IF Format <> fDateTimeDisplay Then
  Begin
    fDateTimeDisplay := Format;
    SetDateTimeFormatString;
    Invalidate;
  End;
End; {SetDateTimeDisplay}


Procedure TDirView.SetDateTimeFormatString;
Begin
  {Create DateTime format string:}
  IF Pos('YYYY', UpperCase(ShortDateFormat)) = 0 Then
  DateTimeFormatStr := ShortDateFormat + 'yy'
  Else
  DateTimeFormatStr := ShortDateFormat ;

  IF fDateTimeDisplay = dtdDateTimeSec Then
  DateTimeFormatStr := DateTimeFormatStr + '  ' + LongTimeFormat
  Else
  IF fDateTimeDisplay = dtdDateTime Then
  DateTimeFormatStr := DateTimeFormatStr + '  ' + ShortTimeFormat
End; {SetDateTimeFormatString}


Procedure TDirView.SetWatchForChanges(Watch : Boolean);
Begin
  IF FWatchForChanges = Watch Then
  Exit;

  FWatchForChanges := Watch;
  IF (csDesigning In ComponentState) Then
  Exit;

  IF Watch Then
    StartWatchThread
  Else
    StopWatchThread;
End; {SetWatchForChanges}


Function  TDirView.FindFileItem(FileName : TFileName) : TListItem;
Var i : Integer;
Begin
  Result := NIL;
  For i := 0 to Pred(Items.Count) Do
  IF FileName = PFileRec(Items[i].Data)^.FileName Then
  Begin
    Result := Items[i];
    Exit;
  End;
End; {FindCaption}



{==========================================================}
{           Events:                                        }
{==========================================================}

(* Event OnColumnClick: *)
procedure TDirView.ColClick(Column: TListColumn);
Var SavedCursor     : TCursor;
    LastSortColumn  : Integer;
    ScrollToFocused : Boolean;

Begin
  ScrollToFocused := Assigned(Selected); { And (Selected.Index >= 0);}

  LastSortColumn := fSortColumn;
  fSortColumn := Column.Index;
  IF fSortColumn > 4 Then
  fSortColumn := 4;

  IF LastSortColumn = fSortColumn Then
  fSortAscending  := Not fSortAscending;

  SavedCursor := Screen.Cursor;
  Screen.Cursor := crHourglass;
  AlphaSort;
  Screen.Cursor := SavedCursor;
  SetColumnImages;
  IF ScrollToFocused And Assigned(Selected) Then
  Selected.MakeVisible(False);

  Inherited ColClick(Column);
End; {Event OnColumnClick}


(* Event OnDblClick: *)
Procedure TDirView.DblClick;
Var DefDir    : String;
    FName     : String;
    Node      : TTreeNode;
    i         : Integer;
    AllowExec : Boolean;

Begin
  Inherited DblClick;

  AllowExec := Not FisRecycleBin;
  IF Assigned(FOnExecFile) Then
  FOnExecFile(Self, Selected, AllowExec);

  IF Not AllowExec Or Not Assigned(Selected) Then
  Exit;

  IF (UpperCase(ExtractFileExt(GetFullFileName(Self.Selected))) = '.LNK') Or
     PFileRec(Self.Selected.Data)^.isDirectory Then
  Begin
    IF PFileRec(Self.Selected.Data)^.isDirectory Then
    FName := GetFullFileName(Self.Selected)
    Else
    FName := ResolveFileShortCut(GetFullFileName(Self.Selected));

    IF DirExists(FName) Then
    Begin
      IF Assigned(fDriveView) Then
      With fDriveView as TDriveView Do
      Begin
        Node := FindNodeToPath(FName);
        IF Not Assigned(Node) Then
        Begin
          ValidateExistingDirectories(GetDriveStatus(FName[1]).RootNode);
          Node := FindNodeToPath(FName);
        End;
        IF Assigned(Node) Then
        Begin
          Directory := FName;
          CenterNode(Selected);
        End;
        Exit;
      End
      Else
      Begin
        Path := FName;
        Exit;
      End;
    End
    Else
    IF Not FileExists(FName) Then
    Exit;
  End;

  GetDir(0, DefDir);
  ChDir(PathName);
  IF Self.SelCount = 1 Then
  ShellExecuteContextCommand(FParentForm.Handle, shcDefault, GetFullFileName(Selected))
  Else
  Begin
    For i := Selected.Index To Items.Count - 1 Do
    IF Items[i].Selected Then
    ShellExecuteContextCommand(FParentForm.Handle, shcDefault, GetFullFileName(Items[i]));
  End;
  ChDir(DefDir);
End; {Event OnDblClick}


(* Event Loaded: *)
Procedure TDirView.Loaded;
Begin
  Inherited Loaded;

  FHeaderhandle := ListView_GetHeader(Self.Handle);
{$IFDEF DRAGDROP}
  FDD.DragDropControl := Self;
{$ENDIF}
  SetSortColumn(SortColumn);
End; {Loaded}


(* Event OnDeletion: *)
procedure TDirView.Delete(Item: TListItem);
Begin
  IF Assigned(Item) And Assigned(Item.Data) Then
  With PFileRec(Item.Data)^ Do
  Begin
    DEC(FFilesSize, Size);
    SetLength(FileName, 0);
    SetLength(TypeName, 0);
    SetLength(DisplayName, 0);
    IF Assigned(PIDL) Then
    PIDL_Free(PIDL);
    Dispose(PFileRec(Item.Data));
    Item.Data := NIL;
  End;
  Inherited Delete(Item);
End; {Delete}


(* Event OnEditing: *)
Function TDirView.CanEdit(Item: TListItem): Boolean;
Begin
  Result := Inherited CanEdit(Item) And
            (fDriveType <> DRIVE_CDROM) And
            Not ReadOnly And
            Not fIsRecycleBin;
  IF Result Then
    FLoadEnabled := False;
End; {CanEdit}


(* Event OnEdited: *)
Procedure TDirView.Edit(Const HItem: TLVItem);
Var Item        : TListItem;
    Node        : TTreeNode;
    isDirectory : Boolean;
    Info        : String;
    i           : Integer;

Begin
  IF Length(HItem.pszText) = 0 Then
  Begin
    LoadEnabled := True;
    Exit;
  End;
  Item := GetItemFromHItem(HItem);

  IF StrContains(coInvalidDosChars, HItem.pszText) Then
  Begin
    Info := coInvalidDosChars;
    For i := Length(Info) DownTo 1 Do
    System.Insert(Space, Info, i);
    Raise EInvalidFileName.Create(ErrorInvalidName + Space + Info);
    LoadEnabled := True;
    Exit;
  End;

  isDirectory := DirExists(GetFullFileName(Item));

  StopWatchThread;
  IF isDirectory And Assigned(fDriveView) Then
  TDriveView(fDriveView).StopWatchThread;

  Try
    IF RenameFile(GetFullFileName(Item), fPath + '\' + HItem.pszText) Then
    Begin
      IF isDirectory And Assigned(fDriveView) Then
      With (fDriveView As TDriveView) Do
      Begin
        Node := FindNodeToPath(GetFullFileName(Item));
        IF Assigned(Node) And Assigned(Node.Parent) Then
          ValidateDirectory(Node.Parent);
      End;


      With GetFileRec(Item.Index)^ Do
      Begin
        IconEmpty   := True;
        FileName    := HItem.pszText;
        DisplayName := FileName;
        FileExt     := UpperCase(Copy(ExtractFileExt(HItem.pszText), 2, 3));
      End;
      GetDisplayData(Item, True);
      Inherited Edit(HItem);
    End
    Else
    Begin
      Item.Caption := GetFileRec(Item.Index)^.FileName;
      Item.Update;
      Raise ERenameFileFailed.Create(ErrorRenameFile + #10 + HItem.pszText);
    End;
  Finally
    LoadEnabled := True;
    IF fWatchForChanges And Not WatchThreadActive Then
    StartWatchThread;

    IF Assigned(fDriveView) Then
    TDriveView(fDriveView).StartWatchThread;
  End;
End; {Edit}


(* Event OnSelectItem *)
Procedure TDirView.InternalOnSelectItem (Sender: TObject; Item: TListItem ; Selected: Boolean);
Begin
  IF Assigned(Item.Data) Then
  Begin
    IF SelCount = 1 Then
      FFilesSelSize := PFileRec(Self.Selected.Data)^.Size
    Else
    Begin
       IF Selected Then
       INC(FFilesSelSize, PFileRec(Item.Data)^.Size)
       Else
       DEC(FFilesSelSize, PFileRec(Item.Data)^.Size);
    End;
  End;
  IF Assigned(fOnSelectItemNew) Then
  fOnSelectItemNew(Sender, Item, Selected);
End; {InternalOnSelectItem}


(* Event OnChange *)
Procedure TDirView.Change(Item: TListItem; Change: Integer);
Begin
  IF not fLoading Then
  Inherited Change(Item, Change);
End; {Change}



(* Event OnKeyDown *)

Procedure TDirView.KeyDown(var Key: Word; Shift: TShiftState);
Begin
  IF (Key = VK_RETURN) Then
  Begin
   IF Shift = [] Then
   DblClick
   Else
   IF Shift = [ssAlt] Then
   DisplayPropertiesMenu;
  End;
  Inherited KeyDown(Key, Shift);
End; {KeyDown}


Procedure TDirView.Notification(AComponent: TComponent; Operation: TOperation);
begin
  Inherited Notification(AComponent, Operation);
  If (Operation = opRemove) and (AComponent = fDriveView) then
  fDriveView := NIL;
end; {Notification}


Procedure TDirView.SetShowMask(Show : Boolean);
Begin
  IF Show <> FShowMask Then
  Begin
    FShowMask := Show;
    SetMask(FMask);
  End;
End; {SetShowMask}


Procedure TDirView.SetColWidthName(W : Integer);
Begin
  IF W <> FColWidthName Then
  Begin
    FColWidthName := W;
    If Columns.Count > 0 Then
    Columns[0].Width := W;
  End;
End; {SetColWidthName}


Procedure TDirView.SetColWidthSize(W : Integer);
Begin
  IF W <> FColWidthSize Then
  Begin
    FColWidthSize := W;
    If Columns.Count > 1 Then
    Columns[1].Width := W;
  End;
End; {SetColWidthSize}


Procedure TDirView.SetColWidthType(W : Integer);
Begin
  IF W <> FColWidthType Then
  Begin
    FColWidthType := W;
    If Columns.Count > 2 Then
    Columns[2].Width := W;
  End;
End; {SetColWidthType}


Procedure TDirView.SetColWidthDate(W : Integer);
Begin
  IF W <> FColWidthDate Then
  Begin
    FColWidthDate := W;
    If Columns.Count > 3 Then
    Columns[3].Width := W;
  End;
End; {SetColWidthDate}


Procedure TDirView.SetColWidthAttr(W : Integer);
Begin
  IF W <> FColWidthAttr Then
  Begin
    FColWidthAttr := W;
    If Columns.Count > 4 Then
    Columns[4].Width := W;
  End;
End; {SetColWidthAttr}


Function TDirView.GetItemFromHItem(Const Item: TLVItem): TListItem;
Begin
  With Item Do
    If (state And LVIF_PARAM) <> 0 then
      Result := Pointer(lParam)
    Else
      Result := Items[iItem];
End; {GetItemFromHItem}


Function TDirView.GetColumnOrder : TColumnOrder;
Begin
  ListView_GetColumnOrderArray(Handle, ColCount, @FColumnOrder);
  Result := FColumnOrder;
End; {GetColumnOrder}


Procedure TDirView.SetColumnOrder(Order: TColumnOrder);
Begin
  FColumnOrder := Order;
  ListView_SetColumnOrderArray(Handle, ColCount, @FColumnOrder);
  Invalidate;
End; {SetColumnOrder}



{$IFDEF DRAGDROP}

Procedure TDirView.WMLButtonDown(var Msg: TWMLButtonDown);
Begin
  GetCursorPos(FStartPos);
  Inherited;
End; {WMLButtonDown}
{$ENDIF}

Procedure TDirView.WMRButtonDown(var Msg: TWMRButtonDown);
Begin
  {$IFDEF DRAGDROP}
  GetCursorPos(FStartPos);
  IF FDD.DragDetectStatus <> ddsDrag Then
  {$ENDIF}
  FContextMenu := True;
  Inherited;
End; {WMRButtonDown}


Procedure TDirView.WMRButtonUp(Var Msg: TWMRButtonUp);
Var WatchThreadOK : Boolean;
    P             : TPoint;

Begin
  IF Assigned(PopupMenu) Then
  PopupMenu.Autopopup := False;
  IF FContextMenu Then
  Begin
    WatchThreadOK := WatchThreadActive;
    IF FUseSystemContextMenu And (SelCount > 0) Then
    Begin
      DisplayContextMenu;
      IF Not WatchThreadOK Then
      ValidateSelectedFiles;
    End
    Else
    Begin
      P.X := Msg.XPos;
      P.Y := Msg.YPos;
      P := ClientToScreen(P);
      IF Assigned(PopupMenu) And Not PopupMenu.AutoPopup Then
        PopupMenu.Popup(P.X, P.Y);
    End;
  End;
  FContextMenu := False;
  Inherited;
End; {WMRButtonUp}



{=================================================================}
{ Drag&Drop handling:                                             }
{=================================================================}
{$IFDEF DRAGDROP}


{Called by TFileDeleteThread, when a file was deleted by the Drag&Drop target window:}
Procedure TDirView.SignalFileDelete(Sender: TObject; Files : TStringList);
Var i : Integer;
Begin
  IF Files.Count > 0 Then
  Begin
    For i := 0 to Files.Count - 1 Do
      ValidateFile(Files[i]);
  End;
End; {SignalFileDelete}




Procedure TDirView.DDDragEnter(DataObj: IDataObject; grfKeyState: Longint; pt: TPoint; var dwEffect: longint; var Accept:boolean);
Var i : Integer;

Begin
  Accept := Accept And (FDriveType <> DRIVE_CDROM) And DirOK;
  IF Accept And (FDD.FileList.Count > 0) And
     (Length(TFDDListItem(FDD.FileList[0]^).Name) > 2) And
     ((TFDDListItem(FDD.FileList[0]^).Name[2] = ':') Or
     (TFDDListItem(FDD.FileList[0]^).Name[2] = '\')) And
     {(Not FisRecycleBin Or (TFDDListItem(FDD.FileList[0]^).MappedName = EmptyStr)) Then}
     (Not FisRecycleBin Or Not FDD.FileNamesAreMapped) Then
  Begin
    FDragDrive := TFDDListItem(FDD.FileList[0]^).Name[1];
    FExeDrag   :=  (FDD.AvailableDropEffects And DropEffect_Link) <> 0;
    IF FExeDrag Then
      For i := 0 To FDD.FileList.Count - 1 Do
        If Not isExecutable(TFDDListItem(FDD.FileList[i]^).Name) Then
        Begin
          FExeDrag := False;
          Break;
        End;
  End
  Else
  Begin
    FDragDrive := #0;
    Accept := False;
  End;

  GetSystemTimeAsFileTime(LastVScrollTime);
  VScrollCount := 0;

  IF Assigned(FOnDDDragEnter) Then
  FOnDDDragEnter(Self, DataObj, grfKeyState, Pt, dwEffect, Accept);
End; {DDDragEnter}


Procedure TDirView.DDDragLeave;
Begin
  IF Assigned(DropTarget) And DragImageList.Dragging Then
  Begin
    DragImageList.HideDragImage;
    DropTarget := NIL;
    Application.ProcessMessages;
    DragImageList.ShowDragImage;
  End
  Else
  DropTarget := NIL;

  IF Assigned(FOnDDDragLeave) Then
  FOnDDDragLeave(Self);
End; {DragLeave}


Procedure TDirView.DDDragOver(grfKeyState: Longint; pt: TPoint; var dwEffect: longint);
Var DropItem : TListItem;
    KnowTime : FILETIME;
    NbPixels : Integer;

Begin
  GetSystemTimeAsFileTime(KnowTime);

  {Set droptarget if traget is directory:}
  Pt := ScreenToClient(Pt);
  DropItem := GetItemAt(Pt.X, PT.Y);

  IF Assigned(DropItem) And Not FisRecycleBin Then
  Begin
    IF PFileRec(DropItem.Data)^.isDirectory  Then
    {DropItem is a directory:}
    Begin
      IF (DropTarget <> DropItem) Then
      Begin
        IF DragImageList.Dragging Then
        Begin
          DragImageList.HideDragImage;
          DropTarget := NIL;
          Application.ProcessMessages;
          DropTarget := DropItem;
          Application.ProcessMessages;
          DragImageList.ShowDragImage;
        End
        Else
        Begin
          DropTarget := NIL;
          DropTarget := DropItem;
        End;
      End;
    End
    Else
    {DropItem is a file:}
    Begin
      IF FDD.OwnerIsSource Then
      dwEffect := DropEffect_None;
      IF Assigned(DropTarget) Then
      Begin
        IF DragImageList.Dragging Then
        Begin
          DragImageList.HideDragImage;
          DropTarget := NIL;
          Application.ProcessMessages;
          DragImageList.ShowDragImage;
        End
        Else
        DropTarget := NIL;
      End;
    End;
  End
  Else
  {DropItem not assigned:}
  Begin
    IF Assigned(DropTarget) Then
    Begin
      IF DragImageList.Dragging Then
      Begin
        DragImageList.HideDragImage;
        DropTarget := NIL;
        Application.ProcessMessages;
        DragImageList.ShowDragImage;
      End
      Else
      DropTarget := NIL;
    End;
    IF FDD.OwnerIsSource Then
    dwEffect := DropEffect_None;
  End;

  {Vertical scrolling, if viewstyle = vsReport:}
  IF (ViewStyle = vsReport) And Assigned(TopItem) Then
  Begin
    NbPixels := Abs((Font.Height));
    IF ((INT64(KnowTime) - INT64(LastVScrollTime)) > DDVScrollDelay) OR
       ((VScrollCount > 5) And ((INT64(KnowTime) - INT64(LastVScrollTime)) > (DDVScrollDelay Div 2))) Then
    Begin
      {Scroll up, if droptarget is topitem:}
      IF (DropItem = TopItem) Or (PT.Y - 3 * nbPixels <= 0) Then
      Begin
        IF (TopItem.Index > 0) Then
        Begin
          If DragImageList.Dragging Then
          DragImageList.HideDragImage;
          Perform(WM_VSCROLL, SB_LINEUP, 0);
          Application.ProcessMessages;
          If DragImageList.Dragging Then
          DragImageList.ShowDragImage;
        End;
        GetSystemTimeAsFileTime(LastVScrollTime);
        INC(VScrollCount);
      End
      Else
      {Scroll down, if next visible item of droptarget is not visible:}
      Begin
        IF PT.Y + 3 * nbPixels > Height Then
        Begin
          Begin
            {IF Assigned(DropItem) And (DropItem.Index < Items.Count - 1) Then}
            Begin
              If DragImageList.Dragging Then
              DragImageList.HideDragImage;
              Perform(WM_VSCROLL, SB_LINEDOWN, 0);
              Application.ProcessMessages;
              If DragImageList.Dragging Then
              DragImageList.ShowDragImage;
            End;
            GetSystemTimeAsFileTime(LastVScrollTime);
            INC(VScrollCount);
          End;
        End
        Else
          VScrollCount := 0;
      End;
    End; {VScrollDelay}
  End;


  {Set dropeffect:}
  IF (grfKeyState And (MK_CONTROL Or MK_SHIFT) = 0) Then
  Begin
    IF FExeDrag And (FPath[1] >= FirstFixedDrive) Then
      dwEffect := DropEffect_Link
    Else
    IF ((FDragDrive = Path[1]) And (dwEffect = DropEffect_Copy) And
       (FDD.AvailableDropEffects and DropEffect_Move <> 0)) Or
       isRecycleBin Then
    dwEffect := DropEffect_Move;
  End;

  IF (Assigned(DropTarget) And PFileRec(DropTarget.Data)^.isRecycleBin) Then
  dwEffect := DropEffect_Move;

  IF Assigned(FOnDDDragOver) Then
  FOnDDDragOver(Self, grfKeyState, Pt, dwEffect);
End; {DDDragOver}


Procedure TDirView.DDDrop(DataObj: IDataObject; grfKeyState: Longint;  pt: TPoint; var dwEffect: longint);

Begin
  IF DragImageList.Dragging Then
  DragImageList.HideDragImage;
  IF dwEffect = DropEffect_None Then
  DropTarget := NIL;
  IF Assigned(FOnDDDrop) Then
  FOnDDDrop(Self, DataObj, grfKeyState, Pt, dwEffect);
End; {DDDrop}


Procedure TDirView.DDQueryContinueDrag(fEscapePressed: BOOL; grfKeyState: Longint; var Result: HResult);
Var P : TPoint;
    ClientP : TPoint;

Begin
  IF Assigned(FOnDDQueryContinueDrag) Then
  FOnDDQueryContinueDrag(Self, fEscapePressed, grfKeyState, Result);

  If FEscapePressed Then
  Begin
    IF DragImageList.Dragging Then
      DragImageList.HideDragImage;
    Exit;
  End;

  IF DragImageList.Dragging Then
  Begin
    GetCursorPos(P);
    {Convert screen coordinates to the parentforms coordinates:}
    ClientP := FParentForm.ScreenToClient(P);
    {Move the drag image to the new position and show it:}
    IF Not CompareMem(@ClientP, @FDragPos, SizeOf(TPoint)) Then
    Begin
      FDragPos := ClientP;
      IF PtInRect(FParentForm.BoundsRect, P) Then
      Begin
        DragImageList.DragMove(ClientP.X, ClientP.Y);
        DragImageList.ShowDragImage;
      End
      Else
        DragImageList.HideDragImage;
    End;
  End;
End; {DDQueryContinueDrag}


Procedure TDirView.DDGiveFeedback(dwEffect: Longint; var Result: HResult);
Begin
  IF Assigned(FOnDDGiveFeedback) Then
  FOnDDGiveFeedback(Self, dwEffect, Result);
End; {DDGiveFeedback}


Procedure TDirView.DDDragDetect(grfKeyState: Longint; DetectStart, Pt: TPoint; DragStatus:TDragDetectStatus);
Var i             : Integer;
    DDResult      : TDragResult;
    WatchThreadOK : Boolean;
    ItemsCount    : Integer;
    SaveCursor    : TCursor;
    P             : TPoint;
    Spot          : TPoint;
    ItemPos       : TPoint;
    Himl          : HImageList;
    BM            : TBitmap;
    DragText      : String;
    FilesCount    : Integer;
    DirsCount     : Integer;

Begin

  IF Assigned(FOnDDDragDetect) Then
  FOnDDDragDetect(Self, grfKeyState, DetectStart, Pt, DragStatus);
  IF (DragStatus = ddsDrag) And (SelCount > 0) Then
  Begin
    SaveCursor := Screen.Cursor;
    Screen.Cursor := crHourGlass;
    IF (SelCount > 100) Or FisRecycleBin Then
      FDD.CompleteFileList := False
    Else
    FDD.CompleteFileList := True;

    ItemsCount := 0;
    FilesCount := 0;
    DirsCount  := 0;

    FDD.FileList.Clear;
    For i := Selected.Index To Items.Count - 1 Do
    Begin
      IF Items[i].Selected Then
      Begin
        If FisRecycleBin Then
          {With PFileRec(Items[i].Data)^ Do}
          FDD.FileList.AddItemEx(PFileRec(Items[i].Data)^.PIDL, GetFullFileName(Items[i]), PFileRec(Items[i].Data)^.DisplayName)
        Else
        FDD.FileList.AddItem(nil,GetFullFileName(Items[i]));
        INC(ItemsCount);
        IF PFileRec(Items[i].Data)^.isDirectory Then
        INC(DirsCount)
        Else
        INC(FilesCount);
        IF ItemsCount >= SelCount Then
        Break;
      End;
    End;

    {Create the dragimage:}
    IF UseDragImages Then
    Begin
      HIml := ListView_CreateDragImage(Handle, Selected.Index, Spot);
      ItemPos := ClientToScreen(Selected.DisplayRect(drBounds).TopLeft);
      IF Himl <> Invalid_Handle_Value Then
      Begin
         DragImageList.Handle := Himl;
         IF SelCount = 1 Then
         Begin
           ItemPos := ClientToScreen(Selected.DisplayRect(drBounds).TopLeft);
           DragImageList.SetDragImage(0, DetectStart.X - ItemPos.X, DetectStart.Y - ItemPos.Y);
         End
         Else
         Begin
           DragImageList.Clear;
           DragImageList.Width := 32;
           DragImageList.Height := 32;
           IF DragImageList.GetResource(rtBitMap, 'DRAGFILES', 0, [lrTransparent], $FFFFFF) Then
           Begin
             BM := TBitmap.Create;
             Try
               Try
                 DragImageList.GetBitmap(0, BM);
                 BM.Canvas.Font.Assign(Self.Font);

                 DragText := EmptyStr;
                 IF FilesCount > 0 Then
                 DragText := Format(TextFiles, [FilesCount]);
                 IF DirsCount > 0 Then
                 Begin
                   IF FilesCount > 0 Then
                   DragText := DragText + ', ';
                   DragText := DragText + Format(TextDirectories, [DirsCount]);
                 End;

                 BM.Width:=33 + BM.Canvas.TextWidth(DragText);
                 BM.TransparentMode := tmAuto;
                 BM.Canvas.TextOut(33, Max(16 - Abs(Canvas.Font.Height), 0), DragText);
                 DragImageList.Clear;
                 DragImageList.Width := BM.Width;
                 DragImageList.Addmasked(BM, BM.Canvas.Pixels[0, 0]);
                 DragImageList.SetDragImage(0, 25, 20);
               Except
               Else
                 If DragImageList.GetResource(rtBitMap, 'DRAGFILES', 0, [lrTransparent], $FFFFFF) Then
                 DragImageList.SetDragImage(0, 25, 20);
               End;
             Finally
               BM.Free;
             End;
           End;
         End;
         P := FParentForm.ScreenToClient(Pt);
         DragImageList.BeginDrag(FParentForm.Handle, P.X, P.Y);
         DragImageList.DragLock(FParentForm.Handle, P.X, P.Y);
         ShowCursor(True);
      End;
    End;

    Screen.Cursor := SaveCursor;
    FContextMenu := False;

    IF FisRecycleBin Then
    FDD.SourceEffects := [deMove]
    Else
    FDD.SourceEffects := [deCopy, deMove, deLink];

    IF DriveType = DRIVE_CDROM Then
    FDD.SourceEffects := FDD.SourceEffects - [deMove]
    Else
    FDD.SourceEffects := FDD.SourceEffects + [deMove];


    WatchThreadOK := WatchThreadActive;
    DropSource := Self;
    GetSystemTimeAsFileTime(DragStartTime);

    {Execute the drag&drop-Operation:}
    DDResult := FDD.Execute;

    {the drag&drop operation is finished, so clean up the used drag image:}
    DragImageList.EndDrag;
    DragImageList.DragUnlock;

    Application.ProcessMessages;
    FContextMenu := False;
    DropTarget := NIL;
    DropSource := NIL;

    IF (DDResult = drMove) And Not WatchThreadOK Then
    Begin
      DragFileList.Clear;
      BuildFileList(DragFileList);
      TFileDeleteThread.Create(DragFileList, MaxWaitTimeOut, SignalFileDelete);
    End;
  End;
End; {(DDDragDetect}


Procedure TDirView.DDProcessDropped(Sender: TObject; grfKeyState: Longint;  pt: TPoint; dwEffect: Longint);
Var KnowTime   : FILETIME;
Begin
  GetSystemTimeAsFileTime(KnowTime);

  IF DirOK And ((INT64(KnowTime) - INT64(DragStartTime)) > DDDragStartDelay) Then
  Begin
    IF Assigned(FOnDDProcessDropped) Then
      FOnDDProcessDropped(Self, grfKeyState, pt, dwEffect);

    IF Assigned(DropTarget) Then
      PerformDragDropFileOperation(AddSlash(PathName) + PFileRec(DropTarget.Data)^.FileName , dwEffect)
    Else
      PerformDragDropFileOperation(PathName, dwEffect);
    IF Assigned(FOnDDExecuted) Then
    FOnDDExecuted(Self, dwEffect);
    FDD.FileList.Clear;
    DropTarget := NIL;
  End;
End; {ProcessDropped}


Procedure TDirView.PerformDragDropFileOperation(TargetPath : String; dwEffect: Integer);
Var i                 : Integer;
    SourcePath        : String;
    SourceFile        : String;
    SourceIsDirectory : Boolean;
    SaveCursor        : TCursor;
    DoFileOperation   : Boolean;
    isRecycleBin      : Boolean;


Begin {DragDropExec}
  SaveCursor    := Screen.Cursor;

  IF FDD.FileList.Count = 0 Then
  Exit;

  IF Not DirExists(TargetPath) Then
  Begin
    Reload(True);
    DDError(DDPathNotFoundError);
    Exit;
  End;

  isRecycleBin := Self.isRecycleBin Or
                  ((Assigned(DropTarget) And Assigned(DropTarget.Data) And PFileRec(DropTarget.Data)^.isRecycleBin));
  IF FDD.FileNamesAreMapped And isRecycleBin Then
  Exit;

  Screen.Cursor := crHourGlass;
  SourceIsDirectory := True;
  SourcePath        := EmptyStr;

  Try
    IF (dwEffect = DropEffect_Copy) Or
       (dwEffect = DropEffect_Move) Then
    Begin
      StopWatchThread;

      IF Assigned(fDriveView) Then
      TDriveView(fDriveView).StopWatchThread;

      IF DropSource is TDirView Then
      TDirView(DropSource).StopWatchThread;

      {Set the source filenames:}
      For i := 0 to FDD.FileList.Count - 1 Do
      Begin
        FFileOperator.OperandFrom.Add(TFDDListItem(FDD.FileList[i]^).Name);
        IF FDD.FileNamesAreMapped Then
        FFileOperator.OperandTo.Add(AddSlash(TargetPath) + TFDDListItem(FDD.FileList[i]^).MappedName);
      End;

      SourcePath := TFDDListItem(FDD.FileList[0]^).Name;

      SourceIsDirectory := DirExists(RemoveSlash(SourcePath));
      IF Not SourceIsDirectory Then
        SourcePath := ExtractFilePath(SourcePath);

      FFileOperator.Flags := [foAllowUndo, foNoConfirmMkDir];

      {Set the target directory or the target filenames:}
      IF FDD.FileNamesAreMapped And Not isRecycleBin Then
        FFileOperator.Flags := FFileOperator.Flags + [foMultiDestFiles]
      Else
      Begin
        FFileOperator.Flags := FFileOperator.Flags - [foMultiDestFiles];
        FFileOperator.OperandTo.Clear;
        FFileOperator.OperandTo.Add(TargetPath);
      End;

      {if the target directory is the recycle bin, then delete the selected files:}
      IF isRecycleBin Then
        FFileOperator.Operation := foDelete
      Else
      Case dwEffect Of
        DropEffect_Copy : FFileOperator.Operation := foCopy;
        DropEffect_Move : FFileOperator.Operation := foMove;
      End;  {Case}


      IF isRecycleBin Then
      Begin
        IF Not ConfirmDelete Then
        FFileOperator.Flags := FFileOperator.Flags + [foNoConfirmation];
      End
      Else
      IF Not ConfirmOverwrite Then
      FFileOperator.Flags := FFileOperator.Flags + [foNoConfirmation];

      DoFileOperation := True;
      IF Assigned(FOnDDFileOperation) Then
      FOnDDFileOperation(Self, dwEffect, SourcePath, TargetPath, DoFileOperation);

      IF DoFileOperation And (FFileOperator.OperandFrom.Count > 0) Then
      Begin
        FFileOperator.Execute;
        IF Assigned(FOnDDFileOperationExecuted) Then
        FOnDDFileOperationExecuted(Self, dwEffect, SourcePath, TargetPath);
        IF FDD.FileNamesAreMapped Then
        FFileOperator.ClearUndo;
      End;
      ReLoad(True);
    End

    Else
    IF (dwEffect = DropEffect_Link) Then
    (* Create Link requested: *)
    Begin
      StopWatchThread;
      For i := 0 to FDD.FileList.Count - 1 Do
      Begin
        SourceFile := TFDDListItem(FDD.FileList[i]^).Name;

        IF Length(SourceFile) = 3 Then
        {Create a link to a drive:}
        SourcePath := Copy(DriveInfo[SourceFile[1]].PrettyName, 4, 255) + ' (' + SourceFile[1] + ')'
        Else
        {Create a link to a file or directory:}
        SourcePath := ExtractFileName(SourceFile);
        IF Not CreateFileShortCut(SourceFile, AddSlash(TargetPath) +  SourcePath + '.lnk',
                                  ExtractFileNameOnly(SourceFile)) Then
          DDError(DDCreateShortCutError);
      End;
      ReLoad(True);
    End;

    IF Assigned(DropSource) And (DropSource is TDirView) And (DropSource <> Self) And (dwEffect = DropEffect_Move) Then
    TDirView(DropSource).ValidateSelectedFiles; 

    IF Assigned(fDriveView) And (SourceIsDirectory or ShowDirectories) Then
    With TDriveView(fDriveView) Do
    Begin
      Try
        ValidateDirectory(FindNodeToPath(TargetPath));
      Except Else End;
      IF ((dwEffect = DropEffect_Move) OR isRecycleBin) Then
      Try
        ValidateDirectory(FindNodeToPath(SourcePath));
      Except Else End;
    End;
  Finally
    FFileOperator.OperandFrom.Clear;
    FFileOperator.OperandTo.Clear;
    IF Not WatchThreadActive Then
    StartWatchThread;
    IF Assigned(fDriveView) Then
    TDriveView(fDriveView).StartWatchThread;
    IF DropSource is TDirView Then
    TDirView(DropSource).StartWatchThread;
    Screen.Cursor := SaveCursor;
  End;
End; {PerformDragDropFileOperation}


Procedure TDirView.DDError(ErrorNo : TDDError);
Begin
  IF Assigned(FOnDDError) Then
  FOnDDError(Self, ErrorNo)
  Else
  Raise EDragDrop.Create('DragDrop Error: ' + IntToStr(Ord(ErrorNo)));
End; {DDError}


Function TDirView.GetCanUndoCopyMove : Boolean;
Begin
  Result := Assigned(FFileOperator) And FFileOperator.CanUndo;
End; {CanUndoCopyMove}


Function TDirView.UndoCopyMove : Boolean;
Var LastTarget : String;
    LastSource : String;

Begin
  Result := False;
  IF FFileOperator.CanUndo Then
  Begin
    Lasttarget := FFileOperator.LastOperandTo[0];
    LastSource := FFileOperator.LastOperandFrom[0];
    Result     := FFileOperator.UndoExecute;

    IF Not WatchthreadActive Then
      Reload(True);

    IF Assigned(FDriveView) Then
    With TDriveView(FDriveView) Do
    Begin
      IF Not WatchThreadActive Or (GetDrive(Selected) <> LastTarget[1]) Then
        ValidateDirectory(GetDriveStatus(LastTarget[1]).RootNode);
        IF LastTarget[1] <> LastSource[1] Then
        ValidateDirectory(GetDriveStatus(LastSource[1]).RootNode);
    End;
  End;
End; {UndoCopyMove}


Procedure TDirView.EmptyClipboard;
Var i : Integer;
Begin
  IF Windows.OpenClipBoard(0) Then
  Begin
    Windows.EmptyClipBoard;
    Windows.CloseClipBoard;
    IF LastClipBoardOperation <> cboNone Then
    For i := 0 To Items.Count - 1 Do
    IF Items[i].Cut Then
    Items[i].Cut := False;
    LastClipBoardOperation := cboNone;
    IF Assigned(FDriveView) Then
      TDriveView(FDriveView).LastPathCut := '';
  End;
End; {EmptyClipBoard}


Function TDirView.CopyToClipBoard : Boolean;
Var i          : Integer;
    SaveCursor : TCursor;
Begin
  SaveCursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;
  Try
    Result := False;
    EmptyClipBoard;
    FDD.FileList.Clear;
    IF SelCount > 0 Then
    Begin
      For i := Selected.Index To Items.Count - 1 Do
      Begin
        IF Items[i].Selected Then
          FDD.FileList.AddItem(nil,GetFullFileName(Items[i]));
      End;
      Result := FDD.CopyToClipBoard;
      LastClipBoardOperation := cboCopy;
    End;
  Finally
    Screen.Cursor := SaveCursor;
  End;
End; {CopyToClipBoard}


Function TDirView.CutToClipBoard : Boolean;
Var i : Integer;
Begin
  Result := False;
  EmptyClipBoard;
  FDD.FileList.Clear;
  IF SelCount > 0 Then
  Begin
    For i := Selected.Index To Items.Count - 1 Do
    Begin
      IF Items[i].Selected Then
      Begin
        FDD.FileList.AddItem(nil,GetFullFileName(Items[i]));
        Items[i].Cut := True;
      End;
    End;
    Result := FDD.CopyToClipBoard;
    LastClipBoardOperation := cboCut;
  End;
End; {CutToClipBoard}


Function TDirView.CanPasteFromClipBoard : Boolean;
Begin
  Result := False;
  IF DirOK And (Path <> '') And (DriveType <> DRIVE_CDROM) And Windows.OpenClipboard(0) Then
  Begin
    Result := IsClipboardFormatAvailable(CF_HDROP);
    Windows.CloseClipBoard;
  End;
End; {CanPasteFromClipBoard}


Function TDirView.PasteFromClipBoard : Boolean;
Begin
  FDD.FileList.Clear;
  Result := False;
  IF CanPasteFromClipBoard And FDD.GetFromClipBoard Then
  Begin
    Case LastClipBoardOperation Of
      cboCopy,
      cboNone: Begin
                 PerformDragDropFileOperation(PathName, DropEffect_Copy);
                 IF Assigned(FOnDDExecuted) Then
                 FOnDDExecuted(Self, DropEffect_Copy);
               End;
      cboCut : Begin
                 PerformDragDropFileOperation(PathName, DropEffect_Move);
                 IF Assigned(FOnDDExecuted) Then
                 FOnDDExecuted(Self, DropEffect_Move);
                 EmptyClipBoard;
               End;
    End;
    Result := True;
  End;
End; {PasteFromClipBoard}


Procedure TDirView.SetTargetPopUpMenu(PopMe : Boolean);
Begin
  IF PopMe <> FTargetPopUpMenu Then
  Begin
    FTargetPopUpMenu := PopMe;
    IF Assigned(FDD) Then
    FDD.TargetPopupMenu := PopMe;
  End;
End; {SetTargetPopUpMenu}

{$ENDIF}

{=================================================================}
{ Shortcut-handling:                                              }
{=================================================================}

Function ResolveFileShortCut(SourceFile : String) : String;
var IUnk   : IUnknown;
    HRes   : HRESULT;        // OLE-Operation Result
    SL     : IShellLink;     // Interface for ShellLink
    PF     : IPersistFile;   // Interface for PersistentFile
    SRec   : TWIN32FINDDATA; // SearchRec of targetfile
    TargetDir : Array[1..Max_Path] Of Char; // Working directory of targetfile
    PSource : WideString;    // Widestring(Source)

Begin
  Result := '';

  IUnk := CreateComObject(CLSID_ShellLink);
  SL := IUnk as IShellLink;
  PF := IUnk as IPersistFile;

  PSource := SourceFile;
  HRes := PF.Load(PWideChar(PSource), STGM_READ);
  IF Succeeded(Hres) Then
  Begin
    HRes := SL.Resolve(Application.Handle, SLR_ANY_MATCH Or SLR_UPDATE{ OR SLR_NO_UI});
    IF Succeeded(Hres) Then
    Begin
      HRes := SL.GetPath(@TargetDir, MAX_PATH, SRec, {SLGP_UNCPRIORITY}{SLGP_SHORTPATH} 0);
      IF Succeeded(Hres) Then
        Result := String(Pchar(@targetDir));
    End;
  End;
  {iUnk._Release;} {<= not required, generates AV}
End; {ResolveShortCut}


Function CreateFileShortCut(SourceFile, Target, DisplayName : String) : Boolean;
var IUnk:      IUnknown;
    Hres:      HRESULT;
    ShellLink: IShellLink;   // Interface to ShellLink
    IPFile:    IPersistFile; // Interface to PersistentFile
    WideStr:   WideString;
    TargetFile : String;

Begin
  Result := False;
  IF Target = '' Then
  TargetFile := SourceFile + '.lnk'
  Else
  TargetFile := Target;

  IUnk := CreateComObject(CLSID_ShellLink);
  ShellLink := IUnk as IShellLink;
  IPFile  := IUnk as IPersistFile;
  with ShellLink do
  begin
    Hres := SetPath(PChar(SourceFile));
    IF Succeeded(Hres) Then
    Hres := SetWorkingDirectory(PChar(ExtractFilePath(SourceFile)));
    IF Succeeded(Hres) Then
    Hres := SetDescription(PChar(DisplayName));
  end;

  WideStr := TargetFile;

  IF Succeeded(Hres) Then
  Begin
    Hres := IPFile.Save(PWChar(WideStr),False);
    IF Succeeded(Hres) Then
    Result := True;
  End;
  {iUnk._Release;} {<= not required, generates AV}
End; {CreateShortCut}


{Sets bitmap bordercolor to systemcolor for 3D-Objects:}
Procedure FixBitmap(hbmp: HBITMAP);
Var Col3D : DWORD;
    Bmp   : TBitmap;

Begin
  Bmp := TBitmap.Create;
  Try
    Bmp.Handle := hbmp;
    Col3D := GetSysColor(COLOR_3DFACE);
    IF Col3D <> DWORD(Bmp.Canvas.Pixels[15, 15]) Then
    Begin
      Bmp.Canvas.Brush.Color := Col3D;
      Bmp.Canvas.FloodFill(15, 15, Bmp.Canvas.Pixels[15, 15], fsSurFace);
{$IFDEF GRAYCOLICON}
      Bmp.Canvas.FloodFill(7, 7, Bmp.Canvas.Pixels[7, 7], fsSurFace);
{$ENDIF}
    End;
  Finally
    bmp.ReleaseHandle;
    bmp.Free;
  End;
End; { FixBitmap }


Function GetIconIndex(Const AFile: String; Attrs: DWORD; Flags: UINT): Integer;
var FileInfo: TSHFileInfo;
begin
  Try
    SHGetFileInfo(PChar(AFile), Attrs, FileInfo, SizeOf(TSHFileInfo),
                  Flags or SHGFI_SYSICONINDEX or SHGFI_USEFILEATTRIBUTES);
    Result := FileInfo.iIcon;
  Except
  Else
    Result := -1;
  End;
end; {GetIconIndex}


Function GetshFileInfo(Const AFile: String; Attrs: DWORD; Flags: UINT): TSHFileInfo;
begin
  SHGetFileInfo(PChar(AFile), Attrs, Result, SizeOf(TSHFileInfo), Flags);
end; {GetshFileInfo}


Function GetShellDisplayName(Const ShellFolder: IShellFolder; IDList: PItemIDList; Flags: DWORD; var Name: string): boolean;
Var Str : TStrRet;

Begin
  Result := TRUE;
  Name := '';
  If ShellFolder.GetDisplayNameOf(IDList, Flags, Str) = NOERROR Then
  Begin
    Case Str.uType of
      STRRET_WSTR:   Name := WideCharToString(Str.pOleStr);
      STRRET_OFFSET: Name := PChar(UINT(IDList) + Str.uOffset);
      STRRET_CSTR:   Name := Str.cStr;
    Else
      Result := FALSE;
    End;
  End
  Else
    Result := FALSE;
End; {GetShellDisplayName}



Function isExecutable(Filename: String): Boolean;
Var FileExt : ExtStr;
Begin
  FileExt := UpperCase(ExtractFileExt(FileName));
  Result := (FileExt = '.EXE') Or (FileExt = '.COM');
End; {isExecuteable}


Function COMCTL32OK : Boolean;
Var VerInfoSize : DWORD;
    Dummy       : DWORD;
    VerInfo     : Pointer;
    FileInfo    : PVSFixedFileInfo;
    FileInfoSize: UINT;

Begin
  Result := False;
  VerInfoSize := GetFileVersionInfoSize('COMCTL32.DLL', Dummy);
  IF VerInfoSize > 0 Then
  Begin
    GetMem(VerInfo, VerInfoSize);
    Try
      IF GetFileVersionInfo(PChar('COMCTL32.DLL'), 0, VerInfoSize, VerInfo) then
      Begin
        IF VerQueryValue(VerInfo, '\', Pointer(FileInfo), FileInfoSize) then
          Result := FileInfo.dwFileVersionMS >= 262214; { COMCTL32 Version >= 4.70  required }
      End;
    Finally
      FreeMem(VerInfo, VerInfoSize);
    End;
  End;
End; {COMCTL32OK}


Function MatchesFileExt(Ext : ExtStr; Const FileExtList : String) : Boolean;
Begin
  Result := (Length(Ext) >= 3) And (Pos(Ext, FileExtList) <> 0);
End; {MatchesFileExt}

{=================================================================}
Initialization
{=================================================================}
  HasExtendedCOMCTL32 := COMCTL32OK;

  {Set language-specific error messages:}
  Case SysLocale.PriLangID Of
    LANG_GERMAN:
    Begin
      ErrorOpenFile   := GERMAN_ErrorOpenFile;
      ErrorRenameFile := GERMAN_ErrorRenameFile;
      ErrorInvalidName:= GERMAN_ErrorInvalidName;
      TextFileExt     := GERMAN_TextFileExt;
      TextFiles       := GERMAN_TextFiles;
      TextDirectories := GERMAN_TextDirectories;
    End;
    LANG_FRENCH:
    Begin
      ErrorOpenFile   := FRENCH_ErrorOpenFile;
      ErrorRenameFile := FRENCH_ErrorRenameFile;
      ErrorInvalidName:= FRENCH_ErrorInvalidName;
      TextFileExt     := FRENCH_TextFileExt;
      TextFiles       := ENGLISH_TextFiles;
      TextDirectories := ENGLISH_TextDirectories;
    End;
    Else
    Begin
      ErrorOpenFile   := ENGLISH_ErrorOpenFile;
      ErrorRenameFile := ENGLISH_ErrorRenameFile;
      ErrorInvalidName:= ENGLISH_ErrorInvalidName;
      TextFileExt     := ENGLISH_TextFileExt;
      TextFiles       := ENGLISH_TextFiles;
      TextDirectories := ENGLISH_TextDirectories;
    End;
  End; {Case}

  SetLength(WinDir, MAX_PATH);
  SetLength(WinDir, GetWindowsDirectory(PChar(WinDir), MAX_PATH));

  SetLength(TempDir, MAX_PATH);
  SetLength(TempDir, GetTempPath(MAX_PATH,PChar(TempDir)));

  UnKnownFileIcon  := GetshFileInfo('',     FILE_ATTRIBUTE_NORMAL,       SHGFI_SYSICONINDEX or SHGFI_USEFILEATTRIBUTES).iIcon;
  DefaultExeIcon   := GetshFileInfo('.COM', FILE_ATTRIBUTE_NORMAL,       SHGFI_SYSICONINDEX or SHGFI_USEFILEATTRIBUTES).iIcon;

  With GetshFileInfo(WinDir, FILE_ATTRIBUTE_NORMAL or FILE_ATTRIBUTE_DIRECTORY, SHGFI_TYPENAME or SHGFI_SYSICONINDEX  or SHGFI_USEFILEATTRIBUTES) Do
  Begin
    StdDirTypeName   := szTypeName;
    StdDirIcon       := iIcon;
  End;
  StdDirSelIcon := GetIconIndex(WinDir, FILE_ATTRIBUTE_NORMAL or FILE_ATTRIBUTE_DIRECTORY, SHGFI_OPENICON);

  LastClipBoardOperation := cboNone;
  DropSource             := NIL;


{=================================================================}
Finalization
{=================================================================}

  SetLength(StdDirTypeName, 0);
  SetLength(ErrorOpenFile, 0);
  SetLength(ErrorRenameFile, 0);
  SetLength(ErrorInvalidName, 0);
  SetLength(TextFileExt, 0);
  SetLength(TextFiles, 0);
  SetLength(WinDir, 0);
End.

